"""
Test the first level model.
"""
import os
import shutil

import numpy as np
import pandas as pd
import pytest
import warnings
from nibabel import Nifti1Image, load
from nibabel.tmpdirs import InTemporaryDirectory
from numpy.testing import (assert_almost_equal, assert_array_almost_equal,
                           assert_array_equal, assert_array_less)

from nilearn._utils.data_gen import (create_fake_bids_dataset,
                                     generate_fake_fmri_data_and_design,
                                     write_fake_fmri_data_and_design)
from nilearn._utils.glm import get_bids_files
from nilearn.glm.contrasts import compute_fixed_effects
from nilearn.glm.first_level import (FirstLevelModel, first_level_from_bids,
                                     mean_scaling, run_glm)
from nilearn.glm.first_level.design_matrix import (
    check_design_matrix, make_first_level_design_matrix)
from nilearn.glm.first_level.first_level import _yule_walker
from nilearn.image import get_data
from nilearn.glm.regression import ARModel, OLSModel
from nilearn.input_data import NiftiMasker

BASEDIR = os.path.dirname(os.path.abspath(__file__))
FUNCFILE = os.path.join(BASEDIR, 'functional.nii.gz')


def test_high_level_glm_one_session():
    shapes, rk = [(7, 8, 9, 15)], 3
    mask, fmri_data, design_matrices = generate_fake_fmri_data_and_design(shapes, rk)

    # Give an unfitted NiftiMasker as mask_img and check that we get an error
    masker = NiftiMasker(mask)
    with pytest.raises(ValueError,
                       match="It seems that NiftiMasker has not been fitted."):
        single_session_model = FirstLevelModel(mask_img=masker).fit(
                fmri_data[0], design_matrices=design_matrices[0])

    # Give a fitted NiftiMasker with a None mask_img_ attribute
    # and check that the masker parameters are overriden by the
    # FirstLevelModel parameters
    masker.fit()
    masker.mask_img_ = None
    with pytest.warns(UserWarning,
                      match="Parameter memory of the masker overriden"):
        single_session_model = FirstLevelModel(mask_img=masker).fit(
                fmri_data[0], design_matrices=design_matrices[0])

    # Give a fitted NiftiMasker
    masker = NiftiMasker(mask)
    masker.fit()
    single_session_model = FirstLevelModel(mask_img=masker).fit(
                    fmri_data[0], design_matrices=design_matrices[0])
    assert single_session_model.masker_ == masker

    # Call with verbose (improve coverage)
    single_session_model = FirstLevelModel(mask_img=None,
                                           verbose=1).fit(
        fmri_data[0], design_matrices=design_matrices[0])

    single_session_model = FirstLevelModel(mask_img=None).fit(
        fmri_data[0], design_matrices=design_matrices[0])
    assert isinstance(single_session_model.masker_.mask_img_,
                      Nifti1Image)

    single_session_model = FirstLevelModel(mask_img=mask).fit(
        fmri_data[0], design_matrices=design_matrices[0])
    z1 = single_session_model.compute_contrast(np.eye(rk)[:1])
    assert isinstance(z1, Nifti1Image)


def test_explicit_fixed_effects():
    """ tests the fixed effects performed manually/explicitly"""
    with InTemporaryDirectory():
        shapes, rk = ((7, 8, 7, 15), (7, 8, 7, 16)), 3
        mask, fmri_data, design_matrices = write_fake_fmri_data_and_design(shapes, rk)
        contrast = np.eye(rk)[1]
        # session 1
        multi_session_model = FirstLevelModel(mask_img=mask).fit(
            fmri_data[0], design_matrices=design_matrices[:1])
        dic1 = multi_session_model.compute_contrast(
            contrast, output_type='all')

        # session 2
        multi_session_model.fit(
            fmri_data[1], design_matrices=design_matrices[1:])
        dic2 = multi_session_model.compute_contrast(
            contrast, output_type='all')

        # fixed effects model
        multi_session_model.fit(
            fmri_data, design_matrices=design_matrices)
        fixed_fx_dic = multi_session_model.compute_contrast(
            contrast, output_type='all')

        # manual version
        contrasts = [dic1['effect_size'], dic2['effect_size']]
        variance = [dic1['effect_variance'], dic2['effect_variance']]
        (
            fixed_fx_contrast,
            fixed_fx_variance,
            fixed_fx_stat,
        ) = compute_fixed_effects(contrasts, variance, mask)

        assert_almost_equal(
            get_data(fixed_fx_contrast),
            get_data(fixed_fx_dic['effect_size']))
        assert_almost_equal(
            get_data(fixed_fx_variance),
            get_data(fixed_fx_dic['effect_variance']))
        assert_almost_equal(
            get_data(fixed_fx_stat), get_data(fixed_fx_dic['stat']))

        # test without mask variable
        (
            fixed_fx_contrast,
            fixed_fx_variance,
            fixed_fx_stat,
        ) = compute_fixed_effects(contrasts, variance)
        assert_almost_equal(
            get_data(fixed_fx_contrast),
            get_data(fixed_fx_dic['effect_size']))
        assert_almost_equal(
            get_data(fixed_fx_variance),
            get_data(fixed_fx_dic['effect_variance']))
        assert_almost_equal(
            get_data(fixed_fx_stat), get_data(fixed_fx_dic['stat']))

        # ensure that using unbalanced effects size and variance images
        # raises an error
        with pytest.raises(ValueError):
            compute_fixed_effects(contrasts * 2, variance, mask)
        del mask, multi_session_model


def test_high_level_glm_with_data():
    with InTemporaryDirectory():
        shapes, rk = ((7, 8, 7, 15), (7, 8, 7, 16)), 3
        mask, fmri_data, design_matrices = write_fake_fmri_data_and_design(shapes, rk)
        multi_session_model = FirstLevelModel(mask_img=None).fit(
            fmri_data, design_matrices=design_matrices)
        n_voxels = get_data(multi_session_model.masker_.mask_img_).sum()
        z_image = multi_session_model.compute_contrast(np.eye(rk)[1])
        assert np.sum(get_data(z_image) != 0) == n_voxels
        assert get_data(z_image).std() < 3.
        # with mask
        multi_session_model = FirstLevelModel(mask_img=mask).fit(
            fmri_data, design_matrices=design_matrices)
        z_image = multi_session_model.compute_contrast(
            np.eye(rk)[:2], output_type='z_score')
        p_value = multi_session_model.compute_contrast(
            np.eye(rk)[:2], output_type='p_value')
        stat_image = multi_session_model.compute_contrast(
            np.eye(rk)[:2], output_type='stat')
        effect_image = multi_session_model.compute_contrast(
            np.eye(rk)[:2], output_type='effect_size')
        variance_image = multi_session_model.compute_contrast(
            np.eye(rk)[:2], output_type='effect_variance')
        assert_array_equal(get_data(z_image) == 0., get_data(load(mask)) == 0.)
        assert (get_data(variance_image)[get_data(load(mask)) > 0] > .001
                ).all()
        all_images = multi_session_model.compute_contrast(
            np.eye(rk)[:2], output_type='all')
        assert_array_equal(get_data(all_images['z_score']), get_data(z_image))
        assert_array_equal(get_data(all_images['p_value']), get_data(p_value))
        assert_array_equal(get_data(all_images['stat']), get_data(stat_image))
        assert_array_equal(get_data(all_images['effect_size']),
                           get_data(effect_image))
        assert_array_equal(get_data(all_images['effect_variance']),
                           get_data(variance_image))
        # Delete objects attached to files to avoid WindowsError when deleting
        # temporary directory (in Windows)
        del (all_images,
             design_matrices,
             effect_image,
             fmri_data,
             mask,
             multi_session_model,
             n_voxels,
             p_value,
             rk,
             shapes,
             stat_image,
             variance_image,
             z_image)


def test_high_level_glm_with_paths():
    shapes, rk = ((7, 8, 7, 15), (7, 8, 7, 14)), 3
    with InTemporaryDirectory():
        mask_file, fmri_files, design_files = write_fake_fmri_data_and_design(shapes, rk)
        multi_session_model = FirstLevelModel(mask_img=None).fit(
            fmri_files, design_matrices=design_files)
        z_image = multi_session_model.compute_contrast(np.eye(rk)[1])
        assert_array_equal(z_image.affine, load(mask_file).affine)
        assert get_data(z_image).std() < 3.
        # Delete objects attached to files to avoid WindowsError when deleting
        # temporary directory (in Windows)
        del z_image, fmri_files, multi_session_model


def test_high_level_glm_null_contrasts():
    # test that contrast computation is resilient to 0 values.
    shapes, rk = ((7, 8, 7, 15), (7, 8, 7, 19)), 3
    mask, fmri_data, design_matrices = generate_fake_fmri_data_and_design(shapes, rk)

    multi_session_model = FirstLevelModel(mask_img=None).fit(
        fmri_data, design_matrices=design_matrices)
    single_session_model = FirstLevelModel(mask_img=None).fit(
        fmri_data[0], design_matrices=design_matrices[0])
    z1 = multi_session_model.compute_contrast([np.eye(rk)[:1],
                                               np.zeros((1, rk))],
                                              output_type='stat')
    z2 = single_session_model.compute_contrast(np.eye(rk)[:1],
                                               output_type='stat')
    np.testing.assert_almost_equal(get_data(z1), get_data(z2))


def test_high_level_glm_different_design_matrices():
    # test that one can estimate a contrast when design matrices are different
    shapes, rk = ((7, 8, 7, 15), (7, 8, 7, 19)), 3
    mask, fmri_data, design_matrices = generate_fake_fmri_data_and_design(shapes, rk)

    # add a column to the second design matrix
    design_matrices[1]['new'] = np.ones((19, 1))

    # Fit a glm with two sessions and design matrices
    multi_session_model = FirstLevelModel(mask_img=mask).fit(
        fmri_data, design_matrices=design_matrices)
    z_joint = multi_session_model.compute_contrast(
        [np.eye(rk)[:1], np.eye(rk + 1)[:1]], output_type='effect_size')
    assert z_joint.shape == (7, 8, 7)

    # compare the estimated effects to seprarately-fitted models
    model1 = FirstLevelModel(mask_img=mask).fit(
        fmri_data[0], design_matrices=design_matrices[0])
    z1 = model1.compute_contrast(np.eye(rk)[:1], output_type='effect_size')
    model2 = FirstLevelModel(mask_img=mask).fit(
        fmri_data[1], design_matrices=design_matrices[1])
    z2 = model2.compute_contrast(np.eye(rk + 1)[:1],
                                 output_type='effect_size')
    assert_almost_equal(get_data(z1) + get_data(z2),
                        2 * get_data(z_joint))


def test_high_level_glm_different_design_matrices_formulas():
    # test that one can estimate a contrast when design matrices are different
    shapes, rk = ((7, 8, 7, 15), (7, 8, 7, 19)), 3
    mask, fmri_data, design_matrices = generate_fake_fmri_data_and_design(shapes, rk)

    # make column names identical
    design_matrices[1].columns = design_matrices[0].columns
    # add a column to the second design matrix
    design_matrices[1]['new'] = np.ones((19, 1))

    # Fit a glm with two sessions and design matrices
    multi_session_model = FirstLevelModel(mask_img=mask).fit(
        fmri_data, design_matrices=design_matrices)

    # Compute contrast with formulas
    cols_formula = tuple(design_matrices[0].columns[:2])
    formula = "%s-%s" % cols_formula
    with pytest.warns(UserWarning, match='One contrast given, assuming it for all 2 runs'):
        z_joint_formula = multi_session_model.compute_contrast(
            formula, output_type='effect_size')


def test_compute_contrast_num_contrasts():

    shapes, rk = ((7, 8, 7, 15), (7, 8, 7, 19), (7, 8, 7, 13)), 3
    mask, fmri_data, design_matrices = generate_fake_fmri_data_and_design(shapes, rk)

    # Fit a glm with 3 sessions and design matrices
    multi_session_model = FirstLevelModel(mask_img=mask).fit(
        fmri_data, design_matrices=design_matrices)

    # raise when n_contrast != n_runs | 1
    with pytest.raises(ValueError):
        multi_session_model.compute_contrast([np.eye(rk)[1]]*2)

    multi_session_model.compute_contrast([np.eye(rk)[1]]*3)
    with pytest.warns(UserWarning, match='One contrast given, assuming it for all 3 runs'):
        multi_session_model.compute_contrast([np.eye(rk)[1]])


def test_run_glm():
    rng = np.random.RandomState(42)
    n, p, q = 33, 80, 10
    X, Y = rng.standard_normal(size=(p, q)), rng.standard_normal(size=(p, n))

    # Ordinary Least Squares case
    labels, results = run_glm(Y, X, 'ols')
    assert_array_equal(labels, np.zeros(n))
    assert list(results.keys()) == [0.0]
    assert results[0.0].theta.shape == (q, n)
    assert_almost_equal(results[0.0].theta.mean(), 0, 1)
    assert_almost_equal(results[0.0].theta.var(), 1. / p, 1)
    assert type(results[labels[0]].model) == OLSModel

    # ar(1) case
    labels, results = run_glm(Y, X, 'ar1')
    assert len(labels) == n
    assert len(results.keys()) > 1
    tmp = sum([val.theta.shape[1] for val in results.values()])
    assert tmp == n
    assert results[labels[0]].model.order == 1
    assert type(results[labels[0]].model) == ARModel

    # ar(3) case
    labels_ar3, results_ar3 = run_glm(Y, X, 'ar3', bins=10)
    assert len(labels_ar3) == n
    assert len(results_ar3.keys()) > 1
    tmp = sum([val.theta.shape[1] for val in results_ar3.values()])
    assert tmp == n
    assert type(results_ar3[labels_ar3[0]].model) == ARModel
    assert results_ar3[labels_ar3[0]].model.order == 3
    assert len(results_ar3[labels_ar3[0]].model.rho) == 3

    # Check correct errors are thrown for nonsense noise model requests
    with pytest.raises(ValueError):
        run_glm(Y, X, 'ar0')
    with pytest.raises(ValueError):
        run_glm(Y, X, 'arfoo')
    with pytest.raises(ValueError):
        run_glm(Y, X, 'arr3')
    with pytest.raises(ValueError):
        run_glm(Y, X, 'ar1.2')
    with pytest.raises(ValueError):
        run_glm(Y, X, 'ar')
    with pytest.raises(ValueError):
        run_glm(Y, X, '3ar')


def test_glm_AR_estimates():
    """Test that Yule-Walker AR fits are correct."""

    n, p, q = 1, 500, 2
    X_orig = np.random.RandomState(2).randn(p, q)
    Y_orig = np.random.RandomState(2).randn(p, n)

    for ar_vals in [[-0.2], [-0.2, -0.5], [-0.2, -0.5, -0.7, -0.3]]:
        ar_order = len(ar_vals)
        ar_arg = 'ar' + str(ar_order)

        X = X_orig.copy()
        Y = Y_orig.copy()

        for idx in range(1, len(Y)):
            for lag in range(ar_order):
                Y[idx] += ar_vals[lag] * Y[idx - 1 - lag]

        # Test using run_glm
        labels, results = run_glm(Y, X, ar_arg, bins=100)
        assert len(labels) == n
        for lab in results.keys():
            ar_estimate = lab.split("_")
            for lag in range(ar_order):
                assert_almost_equal(float(ar_estimate[lag]),
                                    ar_vals[lag], decimal=1)

        # Test using _yule_walker
        yw = _yule_walker(Y.T, ar_order)
        assert_almost_equal(yw[0], ar_vals, decimal=1)

    with pytest.raises(TypeError):
        _yule_walker(Y_orig, 1.2)
    with pytest.raises(ValueError):
        _yule_walker(Y_orig, 0)
    with pytest.raises(ValueError):
        _yule_walker(Y_orig, -2)
    with pytest.raises(TypeError, match='at least 1 dim'):
        _yule_walker(np.array(0.), 2)


def test_scaling():
    """Test the scaling function"""
    rng = np.random.RandomState(42)
    shape = (400, 10)
    u = rng.standard_normal(size=shape)
    mean = 100 * rng.uniform(size=shape[1]) + 1
    Y = u + mean
    Y_, mean_ = mean_scaling(Y)
    assert_almost_equal(Y_.mean(0), 0, 5)
    assert_almost_equal(mean_, mean, 0)
    assert Y.std() > 1


def test_fmri_inputs():
    # Test processing of FMRI inputs
    with InTemporaryDirectory():
        shapes = ((7, 8, 9, 10),)
        mask, FUNCFILE, _ = write_fake_fmri_data_and_design(shapes)
        FUNCFILE = FUNCFILE[0]
        func_img = load(FUNCFILE)
        T = func_img.shape[-1]
        conf = pd.DataFrame([0, 0])
        des = pd.DataFrame(np.ones((T, 1)), columns=[''])
        des_fname = 'design.csv'
        des.to_csv(des_fname)
        events = basic_paradigm()
        for fi in func_img, FUNCFILE:
            for d in des, des_fname:
                FirstLevelModel().fit(fi, design_matrices=d)
                FirstLevelModel(mask_img=None).fit([fi], design_matrices=d)
                FirstLevelModel(mask_img=mask).fit(fi, design_matrices=[d])
                FirstLevelModel(mask_img=mask).fit([fi], design_matrices=[d])
                with pytest.warns(UserWarning, match="If design matrices are supplied"):
                    # test with confounds
                    FirstLevelModel(mask_img=mask).fit([fi], design_matrices=[d],
                                                    confounds=conf)

                # Provide t_r, confounds, and events but no design matrix
                FirstLevelModel(mask_img=mask, t_r=2.0).fit(
                    fi,
                    confounds=pd.DataFrame([0] * 10, columns=['conf']),
                    events=events)

                # Same, but check that an error is raised if there is a
                # mismatch in the dimensions of the inputs
                with pytest.raises(ValueError,
                                   match="Rows in confounds does not match"):
                    FirstLevelModel(mask_img=mask, t_r=2.0).fit(
                            fi, confounds=conf, events=events)

                # test with confounds as numpy array
                FirstLevelModel(mask_img=mask).fit([fi], design_matrices=[d],
                                                   confounds=conf.values)

                FirstLevelModel(mask_img=mask).fit([fi, fi],
                                                   design_matrices=[d, d])
                FirstLevelModel(mask_img=None).fit((fi, fi),
                                                   design_matrices=(d, d))
                with pytest.raises(ValueError):
                    FirstLevelModel(mask_img=None).fit([fi, fi], d)
                with pytest.raises(ValueError):
                    FirstLevelModel(mask_img=None).fit(fi, [d, d])
                # At least paradigms or design have to be given
                with pytest.raises(ValueError):
                    FirstLevelModel(mask_img=None).fit(fi)
                # If paradigms are given then both tr and slice time ref were
                # required
                with pytest.raises(ValueError):
                    FirstLevelModel(mask_img=None).fit(fi, d)
                with pytest.raises(ValueError):
                    FirstLevelModel(mask_img=None, t_r=1.0).fit(fi, d)
                with pytest.raises(ValueError):
                    FirstLevelModel(mask_img=None,
                                    slice_time_ref=0.).fit(fi, d)
            # confounds rows do not match n_scans
            with pytest.raises(ValueError):
                FirstLevelModel(mask_img=None).fit(fi, d, conf)
        # Delete objects attached to files to avoid WindowsError when deleting
        # temporary directory (in Windows)
        del fi, func_img, mask, d, des, FUNCFILE, _


def basic_paradigm():
    conditions = ['c0', 'c0', 'c0', 'c1', 'c1', 'c1', 'c2', 'c2', 'c2']
    onsets = [30, 70, 100, 10, 30, 90, 30, 40, 60]
    durations = 1 * np.ones(9)
    events = pd.DataFrame({'trial_type': conditions,
                           'onset': onsets,
                           'duration': durations})
    return events


def test_first_level_design_creation():
    # Test processing of FMRI inputs
    with InTemporaryDirectory():
        shapes = ((7, 8, 9, 10),)
        mask, FUNCFILE, _ = write_fake_fmri_data_and_design(shapes)
        FUNCFILE = FUNCFILE[0]
        func_img = load(FUNCFILE)
        # basic test based on basic_paradigm and glover hrf
        t_r = 10.0
        slice_time_ref = 0.
        events = basic_paradigm()
        model = FirstLevelModel(t_r, slice_time_ref, mask_img=mask,
                                drift_model='polynomial', drift_order=3)
        model = model.fit(func_img, events)
        frame1, X1, names1 = check_design_matrix(model.design_matrices_[0])
        # check design computation is identical
        n_scans = get_data(func_img).shape[3]
        start_time = slice_time_ref * t_r
        end_time = (n_scans - 1 + slice_time_ref) * t_r
        frame_times = np.linspace(start_time, end_time, n_scans)
        design = make_first_level_design_matrix(frame_times, events,
                                                drift_model='polynomial',
                                                drift_order=3)
        frame2, X2, names2 = check_design_matrix(design)
        assert_array_equal(frame1, frame2)
        assert_array_equal(X1, X2)
        assert_array_equal(names1, names2)
        # Delete objects attached to files to avoid WindowsError when deleting
        # temporary directory (in Windows)
        del FUNCFILE, mask, model, func_img


def test_first_level_glm_computation():
    with InTemporaryDirectory():
        shapes = ((7, 8, 9, 10),)
        mask, FUNCFILE, _ = write_fake_fmri_data_and_design(shapes)
        FUNCFILE = FUNCFILE[0]
        func_img = load(FUNCFILE)
        # basic test based on basic_paradigm and glover hrf
        t_r = 10.0
        slice_time_ref = 0.
        events = basic_paradigm()
        # Ordinary Least Squares case
        model = FirstLevelModel(t_r, slice_time_ref, mask_img=mask,
                                drift_model='polynomial', drift_order=3,
                                minimize_memory=False)
        model = model.fit(func_img, events)

        # Delete objects attached to files to avoid WindowsError when deleting
        # temporary directory (in Windows)
        del mask, FUNCFILE, func_img, model


def test_first_level_glm_computation_with_memory_caching():
    with InTemporaryDirectory():
        shapes = ((7, 8, 9, 10),)
        mask, FUNCFILE, _ = write_fake_fmri_data_and_design(shapes)
        FUNCFILE = FUNCFILE[0]
        func_img = load(FUNCFILE)
        # initialize FirstLevelModel with memory option enabled
        t_r = 10.0
        slice_time_ref = 0.
        events = basic_paradigm()
        # Ordinary Least Squares case
        model = FirstLevelModel(t_r, slice_time_ref, mask_img=mask,
                                drift_model='polynomial', drift_order=3,
                                memory='nilearn_cache', memory_level=1,
                                minimize_memory=False)
        model.fit(func_img, events)
        # Delete objects attached to files to avoid WindowsError when deleting
        # temporary directory (in Windows)
        del mask, func_img, FUNCFILE, model


def test_first_level_contrast_computation():
    with InTemporaryDirectory():
        shapes = ((7, 8, 9, 10),)
        mask, FUNCFILE, _ = write_fake_fmri_data_and_design(shapes)
        FUNCFILE = FUNCFILE[0]
        func_img = load(FUNCFILE)
        # basic test based on basic_paradigm and glover hrf
        t_r = 10.0
        slice_time_ref = 0.
        events = basic_paradigm()
        # Ordinary Least Squares case
        model = FirstLevelModel(t_r, slice_time_ref, mask_img=mask,
                                drift_model='polynomial', drift_order=3,
                                minimize_memory=False)
        c1, c2, cnull = np.eye(7)[0], np.eye(7)[1], np.zeros(7)
        # asking for contrast before model fit gives error
        with pytest.raises(ValueError):
            model.compute_contrast(c1)
        # fit model
        model = model.fit([func_img, func_img], [events, events])
        # Check that an error is raised for invalid contrast_def
        with pytest.raises(ValueError,
                           match="contrast_def must be an array or str or list"):
            model.compute_contrast(37)
        # smoke test for different contrasts in fixed effects
        model.compute_contrast([c1, c2])
        # smoke test for same contrast in fixed effects
        model.compute_contrast([c2, c2])
        # smoke test for contrast that will be repeated
        model.compute_contrast(c2)
        model.compute_contrast(c2, 'F')
        model.compute_contrast(c2, 't', 'z_score')
        model.compute_contrast(c2, 't', 'stat')
        model.compute_contrast(c2, 't', 'p_value')
        model.compute_contrast(c2, None, 'effect_size')
        model.compute_contrast(c2, None, 'effect_variance')
        # formula should work (passing varible name directly)
        model.compute_contrast('c0')
        model.compute_contrast('c1')
        model.compute_contrast('c2')
        # smoke test for one null contrast in group
        model.compute_contrast([c2, cnull])
        # only passing null contrasts should give back a value error
        with pytest.raises(ValueError):
            model.compute_contrast(cnull)
        with pytest.raises(ValueError):
            model.compute_contrast([cnull, cnull])
        # passing wrong parameters
        with pytest.raises(ValueError):
            model.compute_contrast([])
        with pytest.raises(ValueError):
            model.compute_contrast([c1, []])
        with pytest.raises(ValueError):
            model.compute_contrast(c1, '', '')
        with pytest.raises(ValueError):
            model.compute_contrast(c1, '', [])
        # Delete objects attached to files to avoid WindowsError when deleting
        # temporary directory (in Windows)
        del func_img, FUNCFILE, model


def test_first_level_from_bids():
    with InTemporaryDirectory():
        bids_path = create_fake_bids_dataset(n_sub=10, n_ses=2,
                                             tasks=['localizer', 'main'],
                                             n_runs=[1, 3])
        # test arguments are provided correctly
        with pytest.raises(TypeError):
            first_level_from_bids(2, 'main', 'MNI')
        with pytest.raises(ValueError):
            first_level_from_bids('lolo', 'main', 'MNI')
        with pytest.raises(TypeError):
            first_level_from_bids(bids_path, 2, 'MNI')
        with pytest.raises(TypeError):
            first_level_from_bids(bids_path, 'main', 'MNI',
                                         model_init=[])
        with pytest.raises(TypeError,
                           match="space_label must be a string"):
            first_level_from_bids(bids_path, 'main',
                                  space_label=42)

        with pytest.raises(TypeError,
                           match="img_filters must be a list"):
            first_level_from_bids(bids_path, 'main',
                                  img_filters="foo")

        with pytest.raises(TypeError,
                           match="filters in img"):
            first_level_from_bids(bids_path, 'main',
                                  img_filters=[(1, 2)])

        with pytest.raises(ValueError,
                           match="field foo is not a possible filter."):
            first_level_from_bids(bids_path, 'main',
                                  img_filters=[("foo", "bar")])

        # test output is as expected
        models, m_imgs, m_events, m_confounds = first_level_from_bids(
            bids_path, 'main', 'MNI', [('desc', 'preproc')])
        assert len(models) == len(m_imgs)
        assert len(models) == len(m_events)
        assert len(models) == len(m_confounds)
        # test repeated run tag error when run tag is in filenames
        # can arise when desc or space is present and not specified
        with pytest.raises(ValueError):
            first_level_from_bids(
                bids_path, 'main', 'T1w')  # desc not specified
        # test more than one ses file error when run tag is not in filenames
        # can arise when desc or space is present and not specified
        with pytest.raises(ValueError):
            first_level_from_bids(
                bids_path, 'localizer', 'T1w')  # desc not specified
        # test issues with confound files. There should be only one confound
        # file per img. An one per image or None. Case when one is missing
        confound_files = get_bids_files(os.path.join(bids_path, 'derivatives'),
                                        file_tag='desc-confounds_timeseries')
        os.remove(confound_files[-1])
        with pytest.raises(ValueError):
            first_level_from_bids(
                bids_path, 'main', 'MNI')
        # test issues with event files
        events_files = get_bids_files(bids_path, file_tag='events')
        os.remove(events_files[0])
        # one file missing
        with pytest.raises(ValueError):
            first_level_from_bids(bids_path, 'main', 'MNI')
        for f in events_files[1:]:
            os.remove(f)
        # all files missing
        with pytest.raises(ValueError):
            first_level_from_bids(bids_path, 'main', 'MNI')

        # In case different desc and spaces exist and are not selected we
        # fail and ask for more specific information
        shutil.rmtree(os.path.join(bids_path, 'derivatives'))
        # issue if no derivatives folder is present
        with pytest.raises(ValueError):
            first_level_from_bids(
                bids_path, 'main', 'MNI')

        # check runs are not repeated when ses field is not used
        shutil.rmtree(bids_path)
        bids_path = create_fake_bids_dataset(n_sub=10, n_ses=1,
                                             tasks=['localizer', 'main'],
                                             n_runs=[1, 3], no_session=True)
        # test repeated run tag error when run tag is in filenames and not ses
        # can arise when desc or space is present and not specified
        with pytest.raises(ValueError):
            first_level_from_bids(
                bids_path, 'main', 'T1w')  # desc not specified


def test_first_level_with_no_signal_scaling():
    """
    test to ensure that the FirstLevelModel works correctly with a
    signal_scaling==False. In particular, that derived theta are correct for a
    constant design matrix with a single valued fmri image
    """
    shapes, rk = [(3, 1, 1, 2)], 1
    fmri_data = list()
    design_matrices = list()
    design_matrices.append(pd.DataFrame(np.ones((shapes[0][-1], rk)),
                                        columns=list(
                                            'abcdefghijklmnopqrstuvwxyz')[:rk])
                           )
    # Check error with invalid signal_scaling values
    with pytest.raises(ValueError,
                       match="signal_scaling must be"):
        FirstLevelModel(mask_img=False, noise_model='ols',
                        signal_scaling="foo")

    first_level = FirstLevelModel(mask_img=False, noise_model='ols',
                                        signal_scaling=False)
    fmri_data.append(Nifti1Image(np.zeros((1, 1, 1, 2)) + 6, np.eye(4)))

    first_level.fit(fmri_data, design_matrices=design_matrices)
    # trivial test of signal_scaling value
    assert first_level.signal_scaling is False
    # assert that our design matrix has one constant
    assert first_level.design_matrices_[0].equals(
        pd.DataFrame([1.0, 1.0], columns=['a']))
    # assert that we only have one theta as there is only on voxel in our image
    assert first_level.results_[0][0].theta.shape == (1, 1)
    # assert that the theta is equal to the one voxel value
    assert_almost_equal(first_level.results_[0][0].theta[0, 0], 6.0, 2)


def test_first_level_residuals():
    shapes, rk = [(10, 10, 10, 100)], 3
    mask, fmri_data, design_matrices = generate_fake_fmri_data_and_design(shapes, rk)

    for i in range(len(design_matrices)):
        design_matrices[i].iloc[:, 0] = 1

    # Check that voxelwise model attributes cannot be
    # accessed if minimize_memory is set to True
    model = FirstLevelModel(mask_img=mask, minimize_memory=True,
                            noise_model='ols')
    model.fit(fmri_data, design_matrices=design_matrices)

    with pytest.raises(ValueError,
                       match="To access voxelwise attributes"):
        residuals = model.residuals[0]

    model = FirstLevelModel(mask_img=mask, minimize_memory=False,
                            noise_model='ols')

    # Check that trying to access residuals without fitting
    # raises an error
    with pytest.raises(ValueError,
                       match="The model has not been fit yet"):
        residuals = model.residuals[0]

    model.fit(fmri_data, design_matrices=design_matrices)

    # For coverage
    with pytest.raises(ValueError,
                       match="attribute must be one of"):
        model._get_voxelwise_model_attribute("foo", True)
    residuals = model.residuals[0]
    mean_residuals = model.masker_.transform(residuals).mean(0)
    assert_array_almost_equal(mean_residuals, 0)


@pytest.mark.parametrize("shapes", [
    [(10, 10, 10, 25)],
    [(10, 10, 10, 25), (10, 10, 10, 100)],
])
def test_get_voxelwise_attributes_should_return_as_many_as_design_matrices(shapes):
    mask, fmri_data, design_matrices = generate_fake_fmri_data_and_design(shapes)

    for i in range(len(design_matrices)):
        design_matrices[i].iloc[:, 0] = 1

    model = FirstLevelModel(mask_img=mask, minimize_memory=False,
                            noise_model='ols')
    model.fit(fmri_data, design_matrices=design_matrices)

    # Check that length of outputs is the same as the number of design matrices
    assert len(model._get_voxelwise_model_attribute("resid", True)) == len(shapes)


def test_first_level_predictions_r_square():
    shapes, rk = [(10, 10, 10, 25)], 3
    mask, fmri_data, design_matrices = generate_fake_fmri_data_and_design(shapes, rk)

    for i in range(len(design_matrices)):
        design_matrices[i].iloc[:, 0] = 1

    model = FirstLevelModel(mask_img=mask,
                            signal_scaling=False,
                            minimize_memory=False,
                            noise_model='ols')
    model.fit(fmri_data, design_matrices=design_matrices)

    pred = model.predicted[0]
    data = fmri_data[0]
    r_square_3d = model.r_square[0]

    y_predicted = model.masker_.transform(pred)
    y_measured = model.masker_.transform(data)

    assert_almost_equal(np.mean(y_predicted - y_measured), 0)

    r_square_2d = model.masker_.transform(r_square_3d)
    assert_array_less(0., r_square_2d)



import json

import numpy as np
import pytest

from nilearn.plotting.js_plotting_utils import decode
from nilearn.plotting import html_connectome

from .test_js_plotting_utils import check_html


def test_prepare_line():
    e = np.asarray([0, 1, 2, 3], dtype=int)
    n = np.asarray([[0, 1], [0, 2], [2, 3], [8, 9]], dtype=int)
    pe, pn = html_connectome._prepare_line(e, n)
    assert (pn == [0, 1, 0, 0, 2, 0, 2, 3, 0, 8, 9, 0]).all()
    assert(pe == [0, 0, 0, 1, 1, 0, 2, 2, 0, 3, 3, 0]).all()


@pytest.mark.parametrize('node_color,expected_marker_colors', [
    ('cyan', ['#00ffff', '#00ffff', '#00ffff']),
    ('auto', ['#440154', '#20908c', '#fde724']),
    (['cyan', 'red', 'blue'], ['#00ffff', '#ff0000', '#0000ff'])
])
def test_prepare_colors_for_markers(node_color, expected_marker_colors):
    number_of_nodes = 3
    marker_colors = html_connectome._prepare_colors_for_markers(
        node_color,
        number_of_nodes,
    )

    assert marker_colors == expected_marker_colors


def _make_connectome():
    adj = np.diag([1.5, .3, 2.5], 2)
    adj += adj.T
    adj += np.eye(5)

    coord = np.arange(5)
    coord = np.asarray([coord * 10, -coord, coord[::-1]]).T
    return adj, coord


def test_get_connectome():
    adj, coord = _make_connectome()
    connectome = html_connectome._get_connectome(adj, coord)
    con_x = decode(connectome['_con_x'], '<f4')
    expected_x = np.asarray(
        [0, 0, 0,
         0, 20, 0,
         10, 10, 0,
         10, 30, 0,
         20, 0, 0,
         20, 20, 0,
         20, 40, 0,
         30, 10, 0,
         30, 30, 0,
         40, 20, 0,
         40, 40, 0], dtype='<f4')
    assert (con_x == expected_x).all()
    assert {'_con_x', '_con_y', '_con_z', '_con_w'}.issubset(connectome.keys())
    assert (connectome['line_cmin'], connectome['line_cmax']) == (-2.5, 2.5)
    adj[adj == 0] = np.nan
    connectome = html_connectome._get_connectome(adj, coord)
    con_x = decode(connectome['_con_x'], '<f4')
    assert (con_x == expected_x).all()
    assert (connectome['line_cmin'], connectome['line_cmax']) == (-2.5, 2.5)


def test_view_connectome():
    adj, coord = _make_connectome()
    html = html_connectome.view_connectome(adj, coord)
    check_html(html, False, 'connectome-plot')
    html = html_connectome.view_connectome(adj, coord, '85.3%',
                                           title="SOME_TITLE")
    check_html(html, False, 'connectome-plot', title="SOME_TITLE")
    assert "SOME_TITLE" in html.html
    html = html_connectome.view_connectome(adj, coord, '85.3%',
                                           linewidth=8.5, node_size=4.2)
    check_html(html, False, 'connectome-plot', title="Connectome plot")
    html = html_connectome.view_connectome(
        adj, coord, '85.3%', linewidth=8.5, node_size=np.arange(len(coord)))
    check_html(html, False, 'connectome-plot')


def test_view_markers():
    coords = np.arange(12).reshape((4, 3))
    colors = ['r', 'g', 'black', 'white']
    labels = ["red marker", "green marker", 
          "black marker", "white marker"]
    html = html_connectome.view_markers(coords, colors)
    check_html(html, False, 'connectome-plot')
    html = html_connectome.view_markers(coords)
    check_html(html, False, 'connectome-plot')
    html = html_connectome.view_markers(coords, marker_size=15)
    check_html(html, False, 'connectome-plot')
    html = html_connectome.view_markers(
        coords, marker_size=np.arange(len(coords)))
    check_html(html, False, 'connectome-plot')
    html = html_connectome.view_markers(
        coords, marker_size=list(range(len(coords))))
    check_html(html, False, 'connectome-plot')
    html = html_connectome.view_markers(coords, 
                                    marker_size=5.0,
                                    marker_color=colors,
                                    marker_labels=labels)
    labels_dict = {"marker_labels": labels}
    assert json.dumps(labels_dict)[1:-1] in html.html




"""
Transformer for computing seeds signals
----------------------------------------

Mask nifti images by spherical volumes for seed-region analyses
"""
import numpy as np
import warnings
from sklearn import neighbors
from joblib import Memory
from scipy import sparse

from ..image.resampling import coord_transform
from .._utils.niimg_conversions import _safe_get_data
from .._utils import CacheMixin, logger
from .._utils.niimg import img_data_dtype
from .._utils.niimg_conversions import check_niimg_4d, check_niimg_3d
from .._utils.class_inspect import get_params
from .. import image
from .. import masking
from .base_masker import filter_and_extract, BaseMasker


def _apply_mask_and_get_affinity(seeds, niimg, radius, allow_overlap,
                                 mask_img=None):
    '''Utility function to get only the rows which are occupied by sphere at
    given seed locations and the provided radius. Rows are in target_affine and
    target_shape space.

    Parameters
    ----------
    seeds : List of triplets of coordinates in native space
        Seed definitions. List of coordinates of the seeds in the same space
        as target_affine.

    niimg : 3D/4D Niimg-like object
        See http://nilearn.github.io/manipulating_images/input_output.html
        Images to process. It must boil down to a 4D image with scans
        number as last dimension.

    radius : float
        Indicates, in millimeters, the radius for the sphere around the seed.

    allow_overlap: boolean
        If False, a ValueError is raised if VOIs overlap

    mask_img : Niimg-like object, optional
        Mask to apply to regions before extracting signals. If niimg is None,
        mask_img is used as a reference space in which the spheres 'indices are
        placed.

    Returns
    -------
    X : 2D numpy.ndarray
        Signal for each brain voxel in the (masked) niimgs.
        shape: (number of scans, number of voxels)

    A : scipy.sparse.lil_matrix
        Contains the boolean indices for each sphere.
        shape: (number of seeds, number of voxels)

    '''
    seeds = list(seeds)

    # Compute world coordinates of all in-mask voxels.
    if niimg is None:
        mask, affine = masking._load_mask_img(mask_img)
        # Get coordinate for alle voxels inside of mask
        mask_coords = np.asarray(np.nonzero(mask)).T.tolist()
        X = None

    elif mask_img is not None:
        affine = niimg.affine
        mask_img = check_niimg_3d(mask_img)
        mask_img = image.resample_img(mask_img, target_affine=affine,
                                      target_shape=niimg.shape[:3],
                                      interpolation='nearest')
        mask, _ = masking._load_mask_img(mask_img)
        mask_coords = list(zip(*np.where(mask != 0)))

        X = masking._apply_mask_fmri(niimg, mask_img)
    elif niimg is not None:
        affine = niimg.affine
        if np.isnan(np.sum(_safe_get_data(niimg))):
            warnings.warn('The imgs you have fed into fit_transform() contains'
                          ' NaN values which will be converted to zeroes ')
            X = _safe_get_data(niimg, True).reshape([-1, niimg.shape[3]]).T
        else:
            X = _safe_get_data(niimg).reshape([-1, niimg.shape[3]]).T
        mask_coords = list(np.ndindex(niimg.shape[:3]))
    else:
        raise ValueError("Either a niimg or a mask_img must be provided.")

    # For each seed, get coordinates of nearest voxel
    nearests = []
    for sx, sy, sz in seeds:
        nearest = np.round(coord_transform(sx, sy, sz, np.linalg.inv(affine)))
        nearest = nearest.astype(int)
        nearest = (nearest[0], nearest[1], nearest[2])
        try:
            nearests.append(mask_coords.index(nearest))
        except ValueError:
            nearests.append(None)

    mask_coords = np.asarray(list(zip(*mask_coords)))
    mask_coords = coord_transform(mask_coords[0], mask_coords[1],
                                  mask_coords[2], affine)
    mask_coords = np.asarray(mask_coords).T

    clf = neighbors.NearestNeighbors(radius=radius)
    A = clf.fit(mask_coords).radius_neighbors_graph(seeds)
    A = A.tolil()
    for i, nearest in enumerate(nearests):
        if nearest is None:
            continue
        A[i, nearest] = True

    # Include the voxel containing the seed itself if not masked
    mask_coords = mask_coords.astype(int).tolist()
    for i, seed in enumerate(seeds):
        try:
            A[i, mask_coords.index(list(map(int, seed)))] = True
        except ValueError:
            # seed is not in the mask
            pass

    sphere_sizes = np.asarray(A.tocsr().sum(axis=1)).ravel()
    empty_spheres = np.nonzero(sphere_sizes == 0)[0]
    if len(empty_spheres) != 0:
        raise ValueError("These spheres are empty: {}".format(empty_spheres))

    if not allow_overlap:
        if np.any(A.sum(axis=0) >= 2):
            raise ValueError('Overlap detected between spheres')

    return X, A


def _iter_signals_from_spheres(seeds, niimg, radius, allow_overlap,
                               mask_img=None):
    """Utility function to iterate over spheres.

    Parameters
    ----------
    seeds : List of triplets of coordinates in native space
        Seed definitions. List of coordinates of the seeds in the same space
        as the images (typically MNI or TAL).

    niimg : 3D/4D Niimg-like object
        See http://nilearn.github.io/manipulating_images/input_output.html
        Images to process. It must boil down to a 4D image with scans
        number as last dimension.

    radius: float
        Indicates, in millimeters, the radius for the sphere around the seed.

    allow_overlap: boolean
        If False, an error is raised if the maps overlaps (ie at least two
        maps have a non-zero value for the same voxel).

    mask_img : Niimg-like object, optional
        See http://nilearn.github.io/manipulating_images/input_output.html
        Mask to apply to regions before extracting signals.

    """
    X, A = _apply_mask_and_get_affinity(seeds, niimg, radius,
                                        allow_overlap,
                                        mask_img=mask_img)
    for i, row in enumerate(A.rows):
        yield X[:, row]


class _ExtractionFunctor(object):

    func_name = 'nifti_spheres_masker_extractor'

    def __init__(self, seeds_, radius, mask_img, allow_overlap, dtype):
        self.seeds_ = seeds_
        self.radius = radius
        self.mask_img = mask_img
        self.allow_overlap = allow_overlap
        self.dtype = dtype

    def __call__(self, imgs):
        n_seeds = len(self.seeds_)
        imgs = check_niimg_4d(imgs, dtype=self.dtype)

        signals = np.empty((imgs.shape[3], n_seeds), dtype=img_data_dtype(imgs))
        for i, sphere in enumerate(_iter_signals_from_spheres(
                self.seeds_, imgs, self.radius, self.allow_overlap,
                mask_img=self.mask_img)):
            signals[:, i] = np.mean(sphere, axis=1)
        return signals, None


class NiftiSpheresMasker(BaseMasker, CacheMixin):
    """Class for masking of Niimg-like objects using seeds.

    NiftiSpheresMasker is useful when data from given seeds should be
    extracted. Use case: Summarize brain signals from seeds that were
    obtained from prior knowledge.

    Parameters
    ----------
    seeds : List of triplet of coordinates in native space
        Seed definitions. List of coordinates of the seeds in the same space
        as the images (typically MNI or TAL).

    radius : float, optional
        Indicates, in millimeters, the radius for the sphere around the seed.
        Default is None (signal is extracted on a single voxel).

    mask_img : Niimg-like object, optional
        See http://nilearn.github.io/manipulating_images/input_output.html
        Mask to apply to regions before extracting signals.

    allow_overlap : boolean, optional
        If False, an error is raised if the maps overlaps (ie at least two
        maps have a non-zero value for the same voxel). Default=False.

    smoothing_fwhm : float, optional
        If smoothing_fwhm is not None, it gives the full-width half maximum in
        millimeters of the spatial smoothing to apply to the signal.

    standardize : {False, True, 'zscore', 'psc'}, optional
        Strategy to standardize the signal.
        'zscore': the signal is z-scored. Timeseries are shifted
        to zero mean and scaled to unit variance.
        'psc':  Timeseries are shifted to zero mean value and scaled
        to percent signal change (as compared to original mean signal).
        True : the signal is z-scored. Timeseries are shifted
        to zero mean and scaled to unit variance.
        False : Do not standardize the data.
        Default=False.

    standardize_confounds : boolean, optional
        If standardize_confounds is True, the confounds are z-scored:
        their mean is put to 0 and their variance to 1 in the time dimension.
        Default=True.

    high_variance_confounds : boolean, optional
        If True, high variance confounds are computed on provided image with
        :func:`nilearn.image.high_variance_confounds` and default parameters
        and regressed out. Default=False.

    detrend : boolean, optional
        This parameter is passed to signal.clean. Please see the related
        documentation for details. Default=False.

    low_pass : None or float, optional
        This parameter is passed to signal.clean. Please see the related
        documentation for details.

    high_pass : None or float, optional
        This parameter is passed to signal.clean. Please see the related
        documentation for details.

    t_r : float, optional
        This parameter is passed to signal.clean. Please see the related
        documentation for details.

    dtype : {dtype, "auto"}, optional
        Data type toward which the data should be converted. If "auto", the
        data will be converted to int32 if dtype is discrete and float32 if it
        is continuous.

    memory : joblib.Memory or str, optional
        Used to cache the region extraction process.
        By default, no caching is done. If a string is given, it is the
        path to the caching directory.

    memory_level : int, optional
        Aggressiveness of memory caching. The higher the number, the higher
        the number of functions that will be cached. Zero means no caching.
        Default=1.

    verbose : integer, optional
        Indicate the level of verbosity. By default, nothing is printed.
        Default=0.

    See also
    --------
    nilearn.input_data.NiftiMasker

    """
    # memory and memory_level are used by CacheMixin.

    def __init__(self, seeds, radius=None, mask_img=None, allow_overlap=False,
                 smoothing_fwhm=None, standardize=False, standardize_confounds=True,
                 high_variance_confounds=False, detrend=False, low_pass=None,
                 high_pass=None, t_r=None, dtype=None,
                 memory=Memory(location=None, verbose=0), memory_level=1,
                 verbose=0):
        self.seeds = seeds
        self.mask_img = mask_img
        self.radius = radius
        self.allow_overlap = allow_overlap

        # Parameters for _smooth_array
        self.smoothing_fwhm = smoothing_fwhm

        # Parameters for clean()
        self.standardize = standardize
        self.standardize_confounds = standardize_confounds
        self.high_variance_confounds = high_variance_confounds
        self.detrend = detrend
        self.low_pass = low_pass
        self.high_pass = high_pass
        self.t_r = t_r
        self.dtype = dtype

        # Parameters for joblib
        self.memory = memory
        self.memory_level = memory_level

        self.verbose = verbose

    def fit(self, X=None, y=None):
        """Prepare signal extraction from regions.

        All parameters are unused, they are for scikit-learn compatibility.

        """
        if hasattr(self, 'seeds_'):
            return self

        error = ("Seeds must be a list of triplets of coordinates in "
                 "native space.\n")

        if not hasattr(self.seeds, '__iter__'):
            raise ValueError(error + "Given seed list is of type: " +
                             type(self.seeds))

        self.seeds_ = []
        # Check seeds and convert them to lists if needed
        for i, seed in enumerate(self.seeds):
            # Check the type first
            if not hasattr(seed, '__len__'):
                raise ValueError(error + "Seed #%i is not a valid triplet "
                                 "of coordinates. It is of type %s."
                                 % (i, type(seed)))
            # Convert to list because it is easier to process
            if isinstance(seed, np.ndarray):
                seed = seed.tolist()
            else:
                # in case of tuple
                seed = list(seed)

            # Check the length
            if len(seed) != 3:
                raise ValueError(error + "Seed #%i is of length %i "
                                 "instead of 3." % (i, len(seed)))

            self.seeds_.append(seed)

        return self

    def fit_transform(self, imgs, confounds=None, sample_mask=None):
        """Prepare and perform signal extraction"""
        return self.fit().transform(imgs, confounds=confounds,
                                    sample_mask=sample_mask)

    def _check_fitted(self):
        if not hasattr(self, "seeds_"):
            raise ValueError('It seems that %s has not been fitted. '
                             'You must call fit() before calling transform().'
                             % self.__class__.__name__)

    def transform_single_imgs(self, imgs, confounds=None, sample_mask=None):
        """Extract signals from a single 4D niimg.

        Parameters
        ----------
        imgs : 3D/4D Niimg-like object
            See http://nilearn.github.io/manipulating_images/input_output.html
            Images to process. It must boil down to a 4D image with scans
            number as last dimension.

        confounds : CSV file or array-like or pandas DataFrame, optional
            This parameter is passed to signal.clean. Please see the related
            documentation for details.
            shape: (number of scans, number of confounds)

        sample_mask : Any type compatible with numpy-array indexing, optional
            Masks the niimgs along time/fourth dimension to perform scrubbing
            (remove volumes with high motion) and/or non-steady-state volumes.
            This parameter is passed to signal.clean.
            shape: (number of scans - number of volumes removed, )

                .. versionadded:: 0.8.0

        Returns
        -------
        region_signals : 2D numpy.ndarray
            Signal for each sphere.
            shape: (number of scans, number of spheres)

        """
        self._check_fitted()

        params = get_params(NiftiSpheresMasker, self)

        signals, _ = self._cache(
                filter_and_extract,
                ignore=['verbose', 'memory', 'memory_level'])(
            # Images
            imgs, _ExtractionFunctor(self.seeds_, self.radius, self.mask_img,
                                     self.allow_overlap, self.dtype),
            # Pre-processing
            params,
            confounds=confounds,
            sample_mask=sample_mask,
            dtype=self.dtype,
            # Caching
            memory=self.memory,
            memory_level=self.memory_level,
            # kwargs
            verbose=self.verbose)

        return signals

    def inverse_transform(self, region_signals):
        """Compute voxel signals from spheres signals

        Any mask given at initialization is taken into account. Throws an error
        if mask_img==None

        Parameters
        ----------
        region_signals : 2D numpy.ndarray
            Signal for each region.
            shape: (number of scans, number of spheres)

        Returns
        -------
        voxel_signals : nibabel.Nifti1Image
            Signal for each sphere.
            shape: (mask_img, number of scans).

        """
        self._check_fitted()

        logger.log("computing image from signals", verbose=self.verbose)

        if self.mask_img is not None:
            mask = check_niimg_3d(self.mask_img)
        else:
            raise ValueError('Please provide mask_img at initialization to'
                             ' provide a reference for the inverse_transform.')
        _, adjacency = _apply_mask_and_get_affinity(
            self.seeds_, None, self.radius, self.allow_overlap, mask_img=mask)
        adjacency = adjacency.tocsr()
        # Compute overlap scaling for mean signal:
        if self.allow_overlap:
            n_adjacent_spheres = np.asarray(adjacency.sum(axis=0)).ravel()
            scale = 1 / np.maximum(1, n_adjacent_spheres)
            adjacency = adjacency.dot(sparse.diags(scale))

        img = adjacency.T.dot(region_signals.T).T
        return masking.unmask(img, self.mask_img)



import os
import re
import base64
import webbrowser
import tempfile

import numpy as np
import matplotlib
import pytest

from nilearn.plotting import js_plotting_utils
from nilearn import surface
from nilearn.datasets import fetch_surf_fsaverage

try:
    from lxml import etree
    LXML_INSTALLED = True
except ImportError:
    LXML_INSTALLED = False


def _normalize_ws(text):
    return re.sub(r'\s+', ' ', text)


def test_add_js_lib():
    html = js_plotting_utils.get_html_template('surface_plot_template.html')
    cdn = js_plotting_utils.add_js_lib(html, embed_js=False)
    assert "decodeBase64" in cdn
    assert _normalize_ws("""<script
    src="https://ajax.googleapis.com/ajax/libs/jquery/3.3.1/jquery.min.js">
    </script>
    <script src="https://cdn.plot.ly/plotly-gl3d-latest.min.js"></script>
    """) in _normalize_ws(cdn)
    inline = _normalize_ws(js_plotting_utils.add_js_lib(html, embed_js=True))
    assert _normalize_ws("""/*! jQuery v3.3.1 | (c) JS Foundation and other
                            contributors | jquery.org/license */""") in inline
    assert _normalize_ws("""**
                            * plotly.js (gl3d - minified)""") in inline
    assert "decodeBase64" in inline


def check_colors(colors):
    assert len(colors) == 100
    val, cstring = zip(*colors)
    assert np.allclose(np.linspace(0, 1, 100), val, atol=1e-3)
    assert val[0] == 0
    assert val[-1] == 1
    for cs in cstring:
        assert re.match(r'rgb\(\d+, \d+, \d+\)', cs)
    return val, cstring


def test_colorscale_no_threshold():
    cmap = 'jet'
    values = np.linspace(-13, -1.5, 20)
    threshold = None
    colors = js_plotting_utils.colorscale(cmap, values, threshold)
    check_colors(colors['colors'])
    assert (colors['vmin'], colors['vmax']) == (-13, 13)
    assert colors['cmap'].N == 256
    assert (colors['norm'].vmax, colors['norm'].vmin) == (13, -13)
    assert colors['abs_threshold'] is None
    colors = js_plotting_utils.colorscale(cmap, values > 0, .5)


def test_colorscale_threshold_0():
    cmap = 'jet'
    values = np.linspace(-13, -1.5, 20)
    threshold = '0%'
    colors = js_plotting_utils.colorscale(cmap, values, threshold)
    check_colors(colors['colors'])
    assert (colors['vmin'], colors['vmax']) == (-13, 13)
    assert colors['cmap'].N == 256
    assert (colors['norm'].vmax, colors['norm'].vmin) == (13, -13)
    assert colors['abs_threshold'] == 1.5
    assert colors['symmetric_cmap']


def test_colorscale_threshold_99():
    cmap = 'jet'
    values = np.linspace(-13, -1.5, 20)
    threshold = '99%'
    colors = js_plotting_utils.colorscale(cmap, values, threshold)
    check_colors(colors['colors'])
    assert (colors['vmin'], colors['vmax']) == (-13, 13)
    assert colors['cmap'].N == 256
    assert (colors['norm'].vmax, colors['norm'].vmin) == (13, -13)
    assert colors['abs_threshold'] == 13
    assert colors['symmetric_cmap']


def test_colorscale_threshold_50():
    cmap = 'jet'
    values = np.linspace(-13, -1.5, 20)
    threshold = '50%'
    colors = js_plotting_utils.colorscale(cmap, values, threshold)
    val, cstring = check_colors(colors['colors'])
    assert cstring[50] == 'rgb(127, 127, 127)'
    assert (colors['vmin'], colors['vmax']) == (-13, 13)
    assert colors['cmap'].N == 256
    assert (colors['norm'].vmax, colors['norm'].vmin) == (13, -13)
    assert np.allclose(colors['abs_threshold'], 7.55, 2)
    assert colors['symmetric_cmap']


def test_colorscale_absolute_threshold():
    cmap = 'jet'
    values = np.linspace(-13, -1.5, 20)
    threshold = 7.25
    colors = js_plotting_utils.colorscale(cmap, values, threshold)
    val, cstring = check_colors(colors['colors'])
    assert cstring[50] == 'rgb(127, 127, 127)'
    assert (colors['vmin'], colors['vmax']) == (-13, 13)
    assert colors['cmap'].N == 256
    assert (colors['norm'].vmax, colors['norm'].vmin) == (13, -13)
    assert np.allclose(colors['abs_threshold'], 7.25)
    assert colors['symmetric_cmap']


def test_colorscale_asymmetric_cmap():
    cmap = 'jet'
    values = np.arange(15)
    colors = js_plotting_utils.colorscale(cmap, values, symmetric_cmap=False)
    assert (colors['vmin'], colors['vmax']) == (0, 14)
    assert colors['cmap'].N == 256
    assert (colors['norm'].vmax, colors['norm'].vmin) == (14, 0)
    assert not colors['symmetric_cmap']
    values = np.arange(15) + 3
    colors = js_plotting_utils.colorscale(cmap, values, symmetric_cmap=False)
    assert (colors['vmin'], colors['vmax']) == (3, 17)
    assert (colors['norm'].vmax, colors['norm'].vmin) == (17, 3)


def test_colorscale_vmin_vmax():
    cmap = 'jet'
    values = np.arange(15)
    colors = js_plotting_utils.colorscale(cmap, values, vmax=7)
    assert (colors['vmin'], colors['vmax']) == (-7, 7)
    assert colors['cmap'].N == 256
    assert (colors['norm'].vmax, colors['norm'].vmin) == (7, -7)
    assert colors['symmetric_cmap']
    colors = js_plotting_utils.colorscale(
        cmap, values, vmax=7, vmin=-5)
    assert (colors['vmin'], colors['vmax']) == (-7, 7)
    assert colors['cmap'].N == 256
    assert (colors['norm'].vmax, colors['norm'].vmin) == (7, -7)
    assert colors['symmetric_cmap']


def test_colorscale_asymmetric_cmap_vmax():
    cmap = 'jet'
    values = np.arange(15)
    colors = js_plotting_utils.colorscale(cmap, values, vmax=7,
                                          symmetric_cmap=False)
    assert (colors['vmin'], colors['vmax']) == (0, 7)
    assert colors['cmap'].N == 256
    assert (colors['norm'].vmax, colors['norm'].vmin) == (7, 0)
    assert not colors['symmetric_cmap']
    values = np.arange(15) + 3
    colors = js_plotting_utils.colorscale(cmap, values, vmax=7,
                                          symmetric_cmap=False)
    assert (colors['vmin'], colors['vmax']) == (3, 7)
    assert (colors['norm'].vmax, colors['norm'].vmin) == (7, 3)
    colors = js_plotting_utils.colorscale(
        cmap, values, vmax=7, symmetric_cmap=False, vmin=1)
    assert (colors['vmin'], colors['vmax']) == (1, 7)
    assert (colors['norm'].vmax, colors['norm'].vmin) == (7, 1)
    colors = js_plotting_utils.colorscale(
        cmap, values, vmax=10, symmetric_cmap=False, vmin=6, threshold=5)
    assert (colors['vmin'], colors['vmax']) == (0, 10)
    assert (colors['norm'].vmax, colors['norm'].vmin) == (10, 0)
    colors = js_plotting_utils.colorscale(
        cmap, values, vmax=10, symmetric_cmap=False, vmin=None, threshold=5)
    assert (colors['vmin'], colors['vmax']) == (0, 10)
    assert (colors['norm'].vmax, colors['norm'].vmin) == (10, 0)


def test_colorscale_asymmetric_cmap_negative_values():
    cmap = 'jet'
    values = np.linspace(-15, 4)
    with pytest.warns(UserWarning):
        js_plotting_utils.colorscale(cmap,
                                     values, symmetric_cmap=False)

    colors = js_plotting_utils.colorscale(cmap, values, vmax=7,
                                          symmetric_cmap=False)
    assert (colors['vmin'], colors['vmax']) == (-7, 7)
    assert colors['cmap'].N == 256
    assert (colors['norm'].vmax, colors['norm'].vmin) == (7, -7)
    assert colors['symmetric_cmap']


def test_encode():
    for dtype in ['<f4', '<i4', '>f4', '>i4']:
        a = np.arange(10, dtype=dtype)
        encoded = js_plotting_utils.encode(a)
        decoded = base64.b64decode(encoded.encode('utf-8'))
        b = np.frombuffer(decoded, dtype=dtype)
        assert np.allclose(js_plotting_utils.decode(encoded, dtype=dtype), b)
        assert np.allclose(a, b)


def test_mesh_to_plotly():
    fsaverage = fetch_surf_fsaverage()
    coord, triangles = surface.load_surf_mesh(fsaverage['pial_left'])
    plotly = js_plotting_utils.mesh_to_plotly(fsaverage['pial_left'])
    for i, key in enumerate(['_x', '_y', '_z']):
        assert np.allclose(
            js_plotting_utils.decode(plotly[key], '<f4'), coord[:, i])
    for i, key in enumerate(['_i', '_j', '_k']):
        assert np.allclose(
            js_plotting_utils.decode(plotly[key], '<i4'), triangles[:, i])


def check_html(html, check_selects=True, plot_div_id='surface-plot',
               title=None):
    fd, tmpfile = tempfile.mkstemp()
    try:
        os.close(fd)
        html.save_as_html(tmpfile)
        with open(tmpfile) as f:
            saved = f.read()
        # If present, replace Windows line-end '\r\n' with Unix's '\n'
        saved = saved.replace('\r\n', '\n')
        standalone = html.get_standalone().replace('\r\n', '\n')
        assert saved == standalone
    finally:
        os.remove(tmpfile)
    assert "INSERT" not in html.html
    assert html.get_standalone() == html.html
    assert html._repr_html_() == html.get_iframe()
    assert str(html) == html.get_standalone()
    assert '<meta charset="UTF-8" />' in str(html)
    _check_open_in_browser(html)
    resized = html.resize(3, 17)
    assert resized is html
    assert (html.width, html.height) == (3, 17)
    assert 'width="3" height="17"' in html.get_iframe()
    assert 'width="33" height="37"' in html.get_iframe(33, 37)
    if title is not None:
        assert "<title>{}</title>".format(title) in str(html)
    if not LXML_INSTALLED:
        return
    root = etree.HTML(html.html.encode('utf-8'),
                      parser=etree.HTMLParser(huge_tree=True))
    head = root.find('head')
    assert len(head.findall('script')) == 5
    body = root.find('body')
    div = body.find('div')
    assert ('id', plot_div_id) in div.items()
    if not check_selects:
        return
    selects = body.findall('select')
    assert len(selects) == 3
    hemi = selects[0]
    assert ('id', 'select-hemisphere') in hemi.items()
    assert len(hemi.findall('option')) == 2
    kind = selects[1]
    assert ('id', 'select-kind') in kind.items()
    assert len(kind.findall('option')) == 2
    view = selects[2]
    assert ('id', 'select-view') in view.items()
    assert len(view.findall('option')) == 7


def _open_mock(f):
    print('opened {}'.format(f))


def _check_open_in_browser(html):
    wb_open = webbrowser.open
    webbrowser.open = _open_mock
    try:
        html.open_in_browser(temp_file_lifetime=None)
        temp_file = html._temp_file
        assert html._temp_file is not None
        assert os.path.isfile(temp_file)
        html.remove_temp_file()
        assert html._temp_file is None
        assert not os.path.isfile(temp_file)
        html.remove_temp_file()
        html._temp_file = 'aaaaaaaaaaaaaaaaaaaaaa'
        html.remove_temp_file()
    finally:
        webbrowser.open = wb_open
        try:
            os.remove(temp_file)
        except Exception:
            pass


def test_to_color_strings():
    colors = [[0, 0, 1], [1, 0, 0], [.5, .5, .5]]
    as_str = js_plotting_utils.to_color_strings(colors)
    assert as_str == ['#0000ff', '#ff0000', '#7f7f7f']

    colors = [[0, 0, 1, 1], [1, 0, 0, 1], [.5, .5, .5, 0]]
    as_str = js_plotting_utils.to_color_strings(colors)
    assert as_str == ['#0000ff', '#ff0000', '#7f7f7f']

    colors = ['#0000ff', '#ff0000', '#7f7f7f']
    as_str = js_plotting_utils.to_color_strings(colors)
    assert as_str == ['#0000ff', '#ff0000', '#7f7f7f']

    colors = [[0, 0, 1, 1], [1, 0, 0, 1], [.5, .5, .5, 0]]
    as_str = js_plotting_utils.to_color_strings(colors)
    assert as_str == ['#0000ff', '#ff0000', '#7f7f7f']

    colors = ['r', 'green', 'black', 'white']
    as_str = js_plotting_utils.to_color_strings(colors)
    assert as_str == ['#ff0000', '#008000', '#000000', '#ffffff']

    if matplotlib.__version__ < '2':
        return

    colors = ['#0000ffff', '#ff0000ab', '#7f7f7f00']
    as_str = js_plotting_utils.to_color_strings(colors)
    assert as_str == ['#0000ff', '#ff0000', '#7f7f7f']


def test_import_html_document_from_js_plotting():
    from nilearn.plotting.js_plotting_utils import (
        HTMLDocument, set_max_img_views_before_warning)  #noqa



"""
Transformer used to apply basic transformations on MRI data.
"""
# Author: Gael Varoquaux, Alexandre Abraham
# License: simplified BSD

import warnings
import abc

import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin
from joblib import Memory

from .. import masking
from .. import image
from .. import signal
from .. import _utils
from .._utils.cache_mixin import CacheMixin, cache
from .._utils.class_inspect import enclosing_scope_name
from nilearn.image import high_variance_confounds


def filter_and_extract(imgs, extraction_function,
                       parameters,
                       memory_level=0, memory=Memory(location=None),
                       verbose=0,
                       confounds=None,
                       sample_mask=None,
                       copy=True,
                       dtype=None):
    """Extract representative time series using given function.

    Parameters
    ----------
    imgs : 3D/4D Niimg-like object
        Images to be masked. Can be 3-dimensional or 4-dimensional.

    extraction_function : function
        Function used to extract the time series from 4D data. This function
        should take images as argument and returns a tuple containing a 2D
        array with masked signals along with a auxiliary value used if
        returning a second value is needed.
        If any other parameter is needed, a functor or a partial
        function must be provided.

    For all other parameters refer to NiftiMasker documentation

    Returns
    -------
    signals : 2D numpy array
        Signals extracted using the extraction function. It is a scikit-learn
        friendly 2D array with shape n_samples x n_features.

    """
    # Since the calling class can be any *Nifti*Masker, we look for exact type
    if verbose > 0:
        class_name = enclosing_scope_name(stack_level=10)

    # If we have a string (filename), we won't need to copy, as
    # there will be no side effect
    if isinstance(imgs, str):
        copy = False

    if verbose > 0:
        print("[%s] Loading data from %s" % (
            class_name,
            _utils._repr_niimgs(imgs, shorten=False)))
    imgs = _utils.check_niimg(imgs, atleast_4d=True, ensure_ndim=4,
                              dtype=dtype)

    target_shape = parameters.get('target_shape')
    target_affine = parameters.get('target_affine')
    if target_shape is not None or target_affine is not None:
        if verbose > 0:
            print("[%s] Resampling images" % class_name)
        imgs = cache(
            image.resample_img, memory, func_memory_level=2,
            memory_level=memory_level, ignore=['copy'])(
                imgs, interpolation="continuous",
                target_shape=target_shape,
                target_affine=target_affine,
                copy=copy)

    smoothing_fwhm = parameters.get('smoothing_fwhm')
    if smoothing_fwhm is not None:
        if verbose > 0:
            print("[%s] Smoothing images" % class_name)
        imgs = cache(
            image.smooth_img, memory, func_memory_level=2,
            memory_level=memory_level)(
                imgs, parameters['smoothing_fwhm'])

    if verbose > 0:
        print("[%s] Extracting region signals" % class_name)
    region_signals, aux = cache(extraction_function, memory,
                                func_memory_level=2,
                                memory_level=memory_level)(imgs)

    # Temporal
    # --------
    # Detrending (optional)
    # Filtering
    # Confounds removing (from csv file or numpy array)
    # Normalizing
    if verbose > 0:
        print("[%s] Cleaning extracted signals" % class_name)
    runs = parameters.get('runs', None)
    region_signals = cache(
        signal.clean, memory=memory, func_memory_level=2,
        memory_level=memory_level)(
            region_signals,
            detrend=parameters['detrend'],
            standardize=parameters['standardize'],
            standardize_confounds=parameters['standardize_confounds'],
            t_r=parameters['t_r'],
            low_pass=parameters['low_pass'],
            high_pass=parameters['high_pass'],
            confounds=confounds,
            sample_mask=sample_mask,
            runs=runs)

    return region_signals, aux


class BaseMasker(BaseEstimator, TransformerMixin, CacheMixin):
    """Base class for NiftiMaskers
    """

    @abc.abstractmethod
    def transform_single_imgs(self, imgs, confounds=None, sample_mask=None,
                              copy=True):
        """Extract signals from a single 4D niimg.

        Parameters
        ----------
        imgs : 3D/4D Niimg-like object
            See http://nilearn.github.io/manipulating_images/input_output.html
            Images to process. It must boil down to a 4D image with scans
            number as last dimension.

        confounds : CSV file or array-like, optional
            This parameter is passed to signal.clean. Please see the related
            documentation for details.
            shape: (number of scans, number of confounds)

        sample_mask : Any type compatible with numpy-array indexing, optional
            shape: (number of scans - number of volumes removed, )
            Masks the niimgs along time/fourth dimension to perform scrubbing
            (remove volumes with high motion) and/or non-steady-state volumes.
            This parameter is passed to signal.clean.

                .. versionadded:: 0.8.0

        copy : Boolean, optional
            Indicates whether a copy is returned or not. Default=True.

        Returns
        -------
        region_signals : 2D numpy.ndarray
            Signal for each element.
            shape: (number of scans, number of elements)

        """
        raise NotImplementedError()

    def transform(self, imgs, confounds=None, sample_mask=None):
        """Apply mask, spatial and temporal preprocessing

        Parameters
        ----------
        imgs : 3D/4D Niimg-like object
            See http://nilearn.github.io/manipulating_images/input_output.html
            Images to process. It must boil down to a 4D image with scans
            number as last dimension.

        confounds : CSV file or array-like, optional
            This parameter is passed to signal.clean. Please see the related
            documentation for details.
            shape: (number of scans, number of confounds)

        sample_mask : Any type compatible with numpy-array indexing, optional
            shape: (number of scans - number of volumes removed, )
            Masks the niimgs along time/fourth dimension to perform scrubbing
            (remove volumes with high motion) and/or non-steady-state volumes.
            This parameter is passed to signal.clean.

                .. versionadded:: 0.8.0

        Returns
        -------
        region_signals : 2D numpy.ndarray
            Signal for each element.
            shape: (number of scans, number of elements)

        """
        self._check_fitted()

        if confounds is None and not self.high_variance_confounds:
            return self.transform_single_imgs(imgs,
                                              confounds=confounds,
                                              sample_mask=sample_mask)

        # Compute high variance confounds if requested
        all_confounds = []
        if self.high_variance_confounds:
            hv_confounds = self._cache(
                high_variance_confounds)(imgs)
            all_confounds.append(hv_confounds)
        if confounds is not None:
            if isinstance(confounds, list):
                all_confounds += confounds
            else:
                all_confounds.append(confounds)

        return self.transform_single_imgs(imgs, confounds=all_confounds,
                                          sample_mask=sample_mask)

    def fit_transform(self, X, y=None, confounds=None, sample_mask=None,
                      **fit_params):
        """Fit to data, then transform it

        Parameters
        ----------
        X : Niimg-like object
            See http://nilearn.github.io/manipulating_images/input_output.html

        y : numpy array of shape [n_samples], optional
            Target values.

        confounds : list of confounds, optional
            List of confounds (2D arrays or filenames pointing to CSV
            files). Must be of same length than imgs_list.

        sample_mask : list of sample_mask, optional
            List of sample_mask (1D arrays) if scrubbing motion outliers.
            Must be of same length than imgs_list.

                .. versionadded:: 0.8.0

        Returns
        -------
        X_new : numpy array of shape [n_samples, n_features_new]
            Transformed array.

        """
        # non-optimized default implementation; override when a better
        # method is possible for a given clustering algorithm
        if y is None:
            # fit method of arity 1 (unsupervised transformation)
            if self.mask_img is None:
                return self.fit(X, **fit_params
                                ).transform(X, confounds=confounds,
                                            sample_mask=sample_mask)
            else:
                return self.fit(**fit_params).transform(X,
                                                        confounds=confounds,
                                                        sample_mask=sample_mask
                                                        )
        else:
            # fit method of arity 2 (supervised transformation)
            if self.mask_img is None:
                return self.fit(X, y, **fit_params
                                ).transform(X, confounds=confounds,
                                            sample_mask=sample_mask)
            else:
                warnings.warn('[%s.fit] Generation of a mask has been'
                              ' requested (y != None) while a mask has'
                              ' been provided at masker creation. Given mask'
                              ' will be used.' % self.__class__.__name__)
                return self.fit(**fit_params).transform(X, confounds=confounds,
                                                        sample_mask=sample_mask
                                                        )

    def inverse_transform(self, X):
        """ Transform the 2D data matrix back to an image in brain space.

        Parameters
        ----------
        X : Niimg-like object
            See http://nilearn.github.io/manipulating_images/input_output.html

        Returns
        -------
        img : Transformed image in brain space.

        """
        self._check_fitted()
        img = self._cache(masking.unmask)(X, self.mask_img_)
        # Be robust again memmapping that will create read-only arrays in
        # internal structures of the header: remove the memmaped array
        try:
            img._header._structarr = np.array(img._header._structarr).copy()
        except:
            pass
        return img

    def _check_fitted(self):
        if not hasattr(self, "mask_img_"):
            raise ValueError('It seems that %s has not been fitted. '
                             'You must call fit() before calling transform().'
                             % self.__class__.__name__)



import json

import numpy as np
from matplotlib import cm as mpl_cm
from scipy import sparse

from .. import datasets
from . import cm

from .js_plotting_utils import (add_js_lib, mesh_to_plotly,
                                encode, colorscale, get_html_template,
                                to_color_strings)
from nilearn.plotting.html_document import HTMLDocument


class ConnectomeView(HTMLDocument):
    pass


def _encode_coordinates(coords, prefix):
    """
    Transform a 2D-array of 3D data (x, y, z) into a dict of base64 values

    Parameters
    ----------
    coords : ndarray, shape=(n_nodes, 3)
        The coordinates of the nodes in MNI space.

    prefix : str
        Prefix for the key value in the returned dict. Schema is {prefix}{x|y|z}

    Returns
    -------
    coordinates : dict
        Dictionary containing base64 values for each axis
    """
    coordinates = {}

    coords = np.asarray(coords, dtype='<f4')
    marker_x, marker_y, marker_z = coords.T
    for coord, cname in [(marker_x, "x"), (marker_y, "y"), (marker_z, "z")]:
        coordinates["{}{}".format(prefix, cname)] = encode(
            np.asarray(coord, dtype='<f4'))

    return coordinates


def _prepare_line(edges, nodes):
    """prepare a plotly scatter3d line plot so that a set of disconnected edges
    can be drawn as a single line.

    `edges` are values associated with each edge (that get mapped to colors
    through a colorscale). `nodes` are pairs of (source, target) node indices
    for each edge.

    the color of a line segment in plotly is a mixture of the colors associated
    with the points it connects. Moreover, segments that begin or end at a
    point whose value is `null` are not drawn.

    given edges = [eab, ecd, eef] and nodes = [(a, b), (c, d), (e, f)], this
    function returns:

        path_edges: eab eab   0 ecd ecd   0 eef eef   0
        path_nodes:   a   b   0   c   d   0   e   f   0

    moreover the javascript code replaces every third element (the '0' in the
    lists above) with `null`, so only the a-b, c-d, and e-f segments will get
    plotted, and their colors are correct because both their start and end
    points are associated with the same value.
    """
    path_edges = np.zeros(len(edges) * 3, dtype=int)
    path_edges[::3] = edges
    path_edges[1::3] = edges
    path_nodes = np.zeros(len(nodes) * 3, dtype=int)
    path_nodes[::3] = nodes[:, 0]
    path_nodes[1::3] = nodes[:, 1]
    return path_edges, path_nodes


def _prepare_colors_for_markers(marker_color, number_of_nodes):
    """
    Generate "color" and "colorscale" attributes based on `marker_color` mode

    Parameters
    ----------
    marker_color : color or sequence of colors, optional
        Color(s) of the nodes. Default='auto'.

    number_of_nodes : int
        Number of nodes in the view

    Returns
    -------
    markers_colors: list
        List of `number_of_nodes` colors as hexadecimal values
    """
    if isinstance(marker_color, str) and marker_color == 'auto':
        colors = mpl_cm.viridis(np.linspace(0, 1, number_of_nodes))
    elif isinstance(marker_color, str):
        colors = [marker_color] * number_of_nodes
    else:
        colors = marker_color

    return to_color_strings(colors)


def _prepare_lines_metadata(adjacency_matrix, coords, threshold,
                            cmap, symmetric_cmap):
    """
    Generate metadata related to lines for _connectome_view plot

    Parameters
    ----------
    adjacency_matrix : ndarray, shape=(n_nodes, n_nodes)
        The weights of the edges.

    coords : ndarray, shape=(n_nodes, 3)
        The coordinates of the nodes in MNI space.

    threshold : str, number or None, optional
        If None, no thresholding.
        If it is a number only connections of amplitude greater
        than threshold will be shown.
        If it is a string it must finish with a percent sign,
        e.g. "25.3%", and only connections of amplitude above the
        given percentile will be shown.

    cmap : str or matplotlib colormap, optional
        Colormap to use. Default=cm.bwr.

    symmetric_cmap : bool, optional
        Make colormap symmetric (ranging from -vmax to vmax).
        Default=True.

    Returns
    -------
    coordinates : dict
        Dictionary containing base64 values for each axis
    """
    lines_metadata = {}

    adjacency_matrix = np.nan_to_num(adjacency_matrix, copy=True)
    colors = colorscale(
        cmap, adjacency_matrix.ravel(), threshold=threshold,
        symmetric_cmap=symmetric_cmap)
    lines_metadata['line_colorscale'] = colors['colors']
    lines_metadata['line_cmin'] = float(colors['vmin'])
    lines_metadata['line_cmax'] = float(colors['vmax'])
    if threshold is not None:
        adjacency_matrix[
            np.abs(adjacency_matrix) <= colors['abs_threshold']] = 0
    s = sparse.coo_matrix(adjacency_matrix)
    nodes = np.asarray([s.row, s.col], dtype=int).T
    edges = np.arange(len(nodes))
    path_edges, path_nodes = _prepare_line(edges, nodes)
    lines_metadata["_con_w"] = encode(np.asarray(s.data, dtype='<f4')[path_edges])

    line_coords = coords[path_nodes]

    lines_metadata = {
        **lines_metadata,
        **_encode_coordinates(line_coords, prefix="_con_")
    }

    return lines_metadata


def _prepare_markers_metadata(coords, marker_size, marker_color, marker_only):
    markers_coordinates = _encode_coordinates(coords, prefix="_marker_")
    markers_metadata = {
        'markers_only': marker_only,
        **markers_coordinates
    }

    if np.ndim(marker_size) > 0:
        marker_size = np.asarray(marker_size)
    if hasattr(marker_size, 'tolist'):
        marker_size = marker_size.tolist()
    markers_metadata['marker_size'] = marker_size
    markers_metadata['marker_color'] = _prepare_colors_for_markers(
        marker_color,
        len(coords),
    )

    return markers_metadata


def _get_connectome(adjacency_matrix, coords, threshold=None,
                    marker_size=None, marker_color='auto', cmap=cm.cold_hot,
                    symmetric_cmap=True):
    lines_metadata = _prepare_lines_metadata(
        adjacency_matrix,
        coords,
        threshold,
        cmap,
        symmetric_cmap,
    )

    markers_metadata = _prepare_markers_metadata(
        coords,
        marker_size,
        marker_color,
        marker_only=False,
    )

    return {
        **lines_metadata,
        **markers_metadata,
    }


def _make_connectome_html(connectome_info, embed_js=True):
    plot_info = {"connectome": connectome_info}
    mesh = datasets.fetch_surf_fsaverage()
    for hemi in ['pial_left', 'pial_right']:
        plot_info[hemi] = mesh_to_plotly(mesh[hemi])
    as_json = json.dumps(plot_info)
    as_html = get_html_template(
        'connectome_plot_template.html').safe_substitute(
            {'INSERT_CONNECTOME_JSON_HERE': as_json,
             'INSERT_PAGE_TITLE_HERE': (
                 connectome_info["title"] or "Connectome plot")})
    as_html = add_js_lib(as_html, embed_js=embed_js)
    return ConnectomeView(as_html)


def view_connectome(adjacency_matrix, node_coords, edge_threshold=None,
                    edge_cmap=cm.bwr, symmetric_cmap=True,
                    linewidth=6., node_color='auto', node_size=3.,
                    colorbar=True, colorbar_height=.5, colorbar_fontsize=25,
                    title=None, title_fontsize=25):
    """Insert a 3d plot of a connectome into an HTML page.

    Parameters
    ----------
    adjacency_matrix : ndarray, shape=(n_nodes, n_nodes)
        The weights of the edges.

    node_coords : ndarray, shape=(n_nodes, 3)
        The coordinates of the nodes in MNI space.

    node_color : color or sequence of colors, optional
        Color(s) of the nodes. Default='auto'.

    edge_threshold : str, number or None, optional
        If None, no thresholding.
        If it is a number only connections of amplitude greater
        than threshold will be shown.
        If it is a string it must finish with a percent sign,
        e.g. "25.3%", and only connections of amplitude above the
        given percentile will be shown.

    edge_cmap : str or matplotlib colormap, optional
        Colormap to use. Default=cm.bwr.

    symmetric_cmap : bool, optional
        Make colormap symmetric (ranging from -vmax to vmax).
        Default=True.

    linewidth : float, optional
        Width of the lines that show connections. Default=6.0.

    node_size : float, optional
        Size of the markers showing the seeds in pixels.
        Default=3.0.

    colorbar : bool, optional
        Add a colorbar. Default=True.

    colorbar_height : float, optional
        Height of the colorbar, relative to the figure height.
        Default=0.5.

    colorbar_fontsize : int, optional
        Fontsize of the colorbar tick labels. Default=25.

    title : str, optional
        Title for the plot.

    title_fontsize : int, optional
        Fontsize of the title. Default=25.

    Returns
    -------
    ConnectomeView : plot of the connectome.
        It can be saved as an html page or rendered (transparently) by the
        Jupyter notebook. Useful methods are :

        - 'resize' to resize the plot displayed in a Jupyter notebook
        - 'save_as_html' to save the plot to a file
        - 'open_in_browser' to save the plot and open it in a web browser.

    See Also
    --------
    nilearn.plotting.plot_connectome:
        projected views of a connectome in a glass brain.

    nilearn.plotting.view_markers:
        interactive plot of colored markers

    nilearn.plotting.view_surf, nilearn.plotting.view_img_on_surf:
        interactive view of statistical maps or surface atlases on the cortical
        surface.

    """
    node_coords = np.asarray(node_coords)

    connectome_info = _get_connectome(
        adjacency_matrix, node_coords,
        threshold=edge_threshold, cmap=edge_cmap,
        symmetric_cmap=symmetric_cmap, marker_size=node_size,
        marker_color=node_color)
    connectome_info['line_width'] = linewidth
    connectome_info['colorbar'] = colorbar
    connectome_info['cbar_height'] = colorbar_height
    connectome_info['cbar_fontsize'] = colorbar_fontsize
    connectome_info['title'] = title
    connectome_info['title_fontsize'] = title_fontsize
    return _make_connectome_html(connectome_info)


def view_markers(marker_coords, marker_color='auto', marker_size=5.,
                 marker_labels=None, title=None, title_fontsize=25):
    """Insert a 3d plot of markers in a brain into an HTML page.

    Parameters
    ----------
    marker_coords : ndarray, shape=(n_nodes, 3)
        The coordinates of the nodes in MNI space.

    marker_color : ndarray, shape=(n_nodes,), optional
        colors of the markers: list of strings, hex rgb or rgba strings, rgb
        triplets, or rgba triplets (i.e. formats accepted by matplotlib, see
        https://matplotlib.org/users/colors.html#specifying-colors)

    marker_size : float or array-like, optional
        Size of the markers showing the seeds in pixels. Default=5.0.

    marker_labels : list of str, shape=(n_nodes), optional
        Labels for the markers: list of strings

    title : str, optional
        Title for the plot.

    title_fontsize : int, optional
        Fontsize of the title. Default=25.

    Returns
    -------
    ConnectomeView : plot of the markers.
        It can be saved as an html page or rendered (transparently) by the
        Jupyter notebook. Useful methods are :

        - 'resize' to resize the plot displayed in a Jupyter notebook
        - 'save_as_html' to save the plot to a file
        - 'open_in_browser' to save the plot and open it in a web browser.

    See Also
    --------
    nilearn.plotting.plot_connectome:
        projected views of a connectome in a glass brain.

    nilearn.plotting.view_connectome:
        interactive plot of a connectome.

    nilearn.plotting.view_surf, nilearn.plotting.view_img_on_surf:
        interactive view of statistical maps or surface atlases on the cortical
        surface.

    """
    marker_coords = np.asarray(marker_coords)
    if marker_color is None:
        marker_color = ['red' for _ in range(len(marker_coords))]
    connectome_info = _prepare_markers_metadata(
        marker_coords,
        marker_size,
        marker_color,
        marker_only=True,
    )
    if marker_labels is None:
        marker_labels = ['' for _ in range(marker_coords.shape[0])]
    connectome_info['marker_labels'] = marker_labels
    connectome_info['title'] = title
    connectome_info['title_fontsize'] = title_fontsize
    return _make_connectome_html(connectome_info)



"""
Test the neurovault module.
"""
# Author: Jerome Dockes
# License: simplified BSD

import os
import stat
import tempfile
import json
import re
from urllib import parse
import hashlib

import requests
import numpy as np
import pandas as pd
import pytest

from nilearn.datasets import neurovault
from nilearn.image import load_img
from nilearn._utils.data_gen import generate_fake_fmri


def _get_neurovault_data(random_seed=0):
    """Make fake images and collections to mock neurovault in the unit tests.

    Returns two pandas DataFrames: collections and images. Each row contains
    some metadata (e.g. "map_type", "is_thresholded" for images, or
    "number_of_images" for collections) for a single (fake) image or
    collection.

    These two dataframes are like a fake neurovault database and the
    `_neurovault` function, used to simulate responses from the neurovault
    API, uses this data.

    """
    if getattr(_get_neurovault_data, "data", None) is not None:
        return _get_neurovault_data.data
    rng = np.random.RandomState(random_seed)
    n_collections, n_images = 73, 546
    collection_ids = rng.choice(
        np.arange(1000), size=n_collections, replace=False)
    collections = pd.DataFrame({"id": collection_ids})
    image_ids = rng.choice(np.arange(10000), size=n_images, replace=False)
    images = pd.DataFrame({"id": image_ids})
    not_empty = rng.binomial(1, .9, n_collections).astype(bool)
    images["collection_id"] = rng.choice(
        collection_ids[not_empty], size=n_images)
    collection_sizes = images.groupby("collection_id").count()
    collections["true_number_of_images"] = collection_sizes.reindex(
        index=collections["id"].values, fill_value=0).values
    collections["number_of_images"] = collections[
        "true_number_of_images"] + rng.binomial(
            1, .1, n_collections) * rng.randint(0, 100, n_collections)
    images["not_mni"] = rng.binomial(1, .1, size=n_images).astype(bool)
    images["is_valid"] = rng.binomial(1, .1, size=n_images).astype(bool)
    images["is_thresholded"] = rng.binomial(1, .1, size=n_images).astype(bool)
    images["map_type"] = rng.choice(
        ["T map", "Z map", "ROI/mask", "anatomical", "parcellation",
         "something else"], size=n_images, p=[.4, .3, .1, .1, .05, .05])
    images["image_type"] = rng.choice(
        ["statistic_map", "atlas", "other type"],
        size=n_images, p=[.4, .4, .2])
    images["some_key"] = "some_value"
    images[13] = rng.randn(n_images)
    url = "https://neurovault.org/media/images/{}/{}.nii.gz"
    image_names = [
        hashlib.sha1(bytes(img_id)).hexdigest()[:4] for img_id in image_ids]
    images["file"] = [
        url.format(col_id, img_name) for
        (col_id, img_name) in zip(images["collection_id"], image_names)]
    collections.set_index("id", inplace=True, drop=False)
    images.set_index("id", inplace=True, drop=False)
    _get_neurovault_data.data = collections, images
    return collections, images


def _parse_query(query):
    """extract key-value pairs from a url query string

    for example
    "collection=23&someoption&format=json"
      -> {"collection": "23", "someoption": None, "format": "json"}

    """
    parts = [p.split("=") for p in query.split("&")]
    result = {}
    for p in parts:
        if len(p) == 2:
            result[p[0]] = p[1]
        if len(p) == 1:
            result[p[0]] = None
    return result


def _neurovault_collections(parts, query):
    """Mocks the Neurovault API behind the `/api/collections/` path.

    parts: the parts of the URL path after "collections"
     ie [], ["<somecollectionid>"], or ["<somecollectionid>", "images"]

    query: the parsed query string, e.g. {"offset": "15", "limit": "5"}

    returns a dictionary of API results

    See the neurovault API docs for details: https://neurovault.org/api-docs

    """
    if parts:
        return _neurovault_one_collection(parts)
    collections, _ = _get_neurovault_data()
    offset, limit = int(query.get("offset", 0)), int(query.get("limit", 2))
    batch = collections.iloc[
        offset: offset + limit].to_dict(orient="records")
    return {"count": len(collections), "results": batch}


def _neurovault_one_collection(parts):
    """
    Mocks Neurovault API behind the `/api/collections/<somecollectionid>` path.

    parts: parts of the URL path after "collections",
      ie ["<somecollectionid>"] or ["<somecollectionid>", "images"]

    returns a dictionary of API results

    See the neurovault API docs for details: https://neurovault.org/api-docs

    """
    col_id = int(parts[0])
    collections, images = _get_neurovault_data()
    if col_id not in collections.index:
        return {"detail": "Not found."}
    if len(parts) == 1:
        return collections.loc[col_id].to_dict()
    if parts[1] != "images":
        return ""
    col_images = images[images["collection_id"] == col_id]
    return {"count": len(col_images),
            "results": col_images.to_dict(orient="records")}


def _neurovault_images(parts, query):
    """Mocks the Neurovault API behind the `/api/images/` path.

    parts: parts of the URL path after "images",
      ie [] or ["<someimageid>"]

    query: the parsed query string, e.g. {"offset": "15", "limit": "5"}

    returns a dictionary of API results

    See the neurovault API docs for details: https://neurovault.org/api-docs

    """
    if parts:
        return _neurovault_one_image(parts[0])
    _, images = _get_neurovault_data()
    offset, limit = int(query.get("offset", 0)), int(query.get("limit", 2))
    batch = images.iloc[offset: offset + limit].to_dict(orient="records")
    return {"count": len(images), "results": batch}


def _neurovault_one_image(img_id):
    """Mocks the Neurovault API behind the `/api/images/<someimageid>` path.

    returns a dictionary of API results

    See the neurovault API docs for details: https://neurovault.org/api-docs

    """
    img_id = int(img_id)
    _, images = _get_neurovault_data()
    if img_id not in images.index:
        return {"detail": "Not found."}
    return images.loc[img_id].to_dict()


def _neurovault_file(parts, query):
    """Mocks the Neurovault API behind the `/media/images/` path."""
    return generate_fake_fmri(length=1)[0]


class _NumpyJsonEncoder(json.JSONEncoder):
    """A json encoder that can handle numpy objects"""
    def default(self, obj):
        if hasattr(obj, "tolist"):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def _neurovault(match, request):
    """Mock response content from the Neurovault API.

    The fake data used to generate responses is provided by
    `_get_neurovault_data`.

    See the neurovault API docs for details on the queries and corresponding
    responses: https://neurovault.org/api-docs

    """
    handlers = {
        "media": {"images": _neurovault_file},
        "api": {
            "collections": _neurovault_collections,
            "images": _neurovault_images,
        }
    }
    info = parse.urlparse(request.url)
    parts = list(filter(bool, info.path.split("/")))
    endpoint, section = parts[0], parts[1]

    result = handlers[endpoint][section](parts[2:], _parse_query(info.query))
    should_jsonify_response = endpoint == "api"
    return (
        json.dumps(result, cls=_NumpyJsonEncoder).encode("UTF-8")
        if should_jsonify_response
        else result
    )


@pytest.fixture(autouse=True)
def neurovault_mocker(request_mocker):
    request_mocker.url_mapping["*neurovault.org*"] = _neurovault


def test_remove_none_strings():
    info = {'a': 'None / Other',
            'b': '',
            'c': 'N/A',
            'd': None,
            'e': 0,
            'f': 'a',
            'g': 'Name'}
    assert (neurovault._remove_none_strings(info) ==
                 {'a': None,
                  'b': None,
                  'c': None,
                  'd': None,
                  'e': 0,
                  'f': 'a',
                  'g': 'Name'})


def test_append_filters_to_query():
    query = neurovault._append_filters_to_query(
        neurovault._NEUROVAULT_COLLECTIONS_URL,
        {'DOI': 17})
    assert (
        query == 'http://neurovault.org/api/collections/?DOI=17')
    query = neurovault._append_filters_to_query(
        neurovault._NEUROVAULT_COLLECTIONS_URL,
        {'id': 40})
    assert query == 'http://neurovault.org/api/collections/40'


def test_get_batch():
    batch = neurovault._get_batch(neurovault._NEUROVAULT_COLLECTIONS_URL)
    assert('results' in batch)
    assert('count' in batch)
    pytest.raises(requests.RequestException, neurovault._get_batch, 'http://')
    with tempfile.TemporaryDirectory() as temp_dir:
        with open(os.path.join(temp_dir, 'test_nv.txt'), 'w'):
            pass
        pytest.raises(ValueError, neurovault._get_batch, 'file://{0}'.format(
            os.path.join(temp_dir, 'test_nv.txt')))
    no_results_url = ('https://eutils.ncbi.nlm.nih.gov/entrez/eutils/'
                      'esearch.fcgi?db=pmc&retmode=json&term=fmri')
    pytest.raises(ValueError, neurovault._get_batch, no_results_url)


def test_scroll_server_results():
    result = list(neurovault._scroll_server_results(
        neurovault._NEUROVAULT_COLLECTIONS_URL, max_results=6, batch_size=3))
    assert len(result) == 6
    result = list(neurovault._scroll_server_results(
        neurovault._NEUROVAULT_COLLECTIONS_URL, max_results=3,
        local_filter=lambda r: False))
    assert len(result) == 0
    no_results = neurovault._scroll_server_results(
        'http://BAD_URL', max_results=3,
        local_filter=lambda r: True)
    next(no_results)


def test_is_null():
    is_null = neurovault.IsNull()
    assert is_null != 'a'
    assert not is_null != ''
    assert 'a' != is_null
    assert not '' != is_null
    assert not is_null == 'a'
    assert is_null == ''
    assert not 'a' == is_null
    assert '' == is_null
    assert str(is_null) == 'IsNull()'


def test_not_null():
    not_null = neurovault.NotNull()
    assert not_null == 'a'
    assert not not_null == ''
    assert 'a' == not_null
    assert not '' == not_null
    assert not not_null != 'a'
    assert not_null != ''
    assert not 'a' != not_null
    assert '' != not_null
    assert str(not_null) == 'NotNull()'


def test_not_equal():
    not_equal = neurovault.NotEqual('a')
    assert not_equal == 'b'
    assert not_equal == 1
    assert not not_equal == 'a'
    assert 'b' == not_equal
    assert 1 == not_equal
    assert not 'a' == not_equal
    assert not not_equal != 'b'
    assert not not_equal != 1
    assert not_equal != 'a'
    assert not 'b' != not_equal
    assert not 1 != not_equal
    assert 'a' != not_equal
    assert str(not_equal) == "NotEqual('a')"


def test_order_comp():
    geq = neurovault.GreaterOrEqual('2016-07-12T11:29:12.263046Z')
    assert '2016-08-12T11:29:12.263046Z' == geq
    assert '2016-06-12T11:29:12.263046Z' != geq
    assert str(geq) == "GreaterOrEqual('2016-07-12T11:29:12.263046Z')"
    gt = neurovault.GreaterThan('abc')
    assert not gt == 'abc'
    assert gt == 'abd'
    assert str(gt) == "GreaterThan('abc')"
    lt = neurovault.LessThan(7)
    assert not 7 == lt
    assert not 5 != lt
    assert not lt == 'a'
    assert str(lt) == 'LessThan(7)'
    leq = neurovault.LessOrEqual(4.5)
    assert 4.4 == leq
    assert not 4.6 == leq
    assert str(leq) == 'LessOrEqual(4.5)'


def test_is_in():
    is_in = neurovault.IsIn(0, 1)
    assert is_in == 0
    assert not is_in == 2
    assert 0 == is_in
    assert not 2 == is_in
    assert not is_in != 0
    assert is_in != 2
    assert not 0 != is_in
    assert 2 != is_in
    assert str(is_in) == 'IsIn(0, 1)'
    countable = neurovault.IsIn(*range(11))
    assert 7 == countable
    assert not countable == 12


def test_not_in():
    not_in = neurovault.NotIn(0, 1)
    assert not_in != 0
    assert not not_in != 2
    assert 0 != not_in
    assert not 2 != not_in
    assert not not_in == 0
    assert not_in == 2
    assert not 0 == not_in
    assert 2 == not_in
    assert str(not_in) == 'NotIn(0, 1)'


def test_contains():
    contains = neurovault.Contains('a', 0)
    assert not contains == 10
    assert contains == ['b', 1, 'a', 0]
    assert ['b', 1, 'a', 0] == contains
    assert contains != ['b', 1, 0]
    assert ['b', 1, 'a'] != contains
    assert not contains != ['b', 1, 'a', 0]
    assert not ['b', 1, 'a', 0] != contains
    assert not contains == ['b', 1, 0]
    assert not ['b', 1, 'a'] == contains
    assert str(contains) == "Contains('a', 0)"
    contains = neurovault.Contains('house', 'face')
    assert 'face vs house' == contains
    assert not 'smiling face vs frowning face' == contains


def test_not_contains():
    not_contains = neurovault.NotContains('ab')
    assert None != not_contains
    assert not_contains == 'a_b'
    assert 'bcd' == not_contains
    assert not_contains != '_abcd'
    assert '_abcd' != not_contains
    assert not not_contains != 'a_b'
    assert not 'bcd' != not_contains
    assert not not_contains == '_abcd'
    assert not '_abcd' == not_contains
    assert str(not_contains) == "NotContains('ab',)"


def test_pattern():
    # Python std lib doc poker hand example
    pattern_0 = neurovault.Pattern(r'[0-9akqj]{5}$')
    assert str(pattern_0) == "Pattern(pattern='[0-9akqj]{5}$', flags=0)"
    pattern_1 = neurovault.Pattern(r'[0-9akqj]{5}$', re.I)
    assert pattern_0 == 'ak05q'
    assert not pattern_0 == 'Ak05q'
    assert not pattern_0 == 'ak05e'
    assert pattern_1 == 'ak05q'
    assert pattern_1 == 'Ak05q'
    assert not pattern_1 == 'ak05e'
    assert not pattern_0 != 'ak05q'
    assert pattern_0 != 'Ak05q'
    assert pattern_0 != 'ak05e'
    assert not pattern_1 != 'ak05q'
    assert not pattern_1 != 'Ak05q'
    assert pattern_1 != 'ak05e'

    assert 'ak05q' == pattern_0
    assert not 'Ak05q' == pattern_0
    assert not 'ak05e' == pattern_0
    assert 'ak05q' == pattern_1
    assert 'Ak05q' == pattern_1
    assert not 'ak05e' == pattern_1
    assert not 'ak05q' != pattern_0
    assert 'Ak05q' != pattern_0
    assert 'ak05e' != pattern_0
    assert not 'ak05q' != pattern_1
    assert not 'Ak05q' != pattern_1
    assert 'ak05e' != pattern_1


def test_result_filter():
    filter_0 = neurovault.ResultFilter(query_terms={'a': 0},
                                       callable_filter=lambda d: len(d) < 5,
                                       b=1)
    assert str(filter_0) == 'ResultFilter'
    assert filter_0['a'] == 0
    assert filter_0({'a': 0, 'b': 1, 'c': 2})
    assert not filter_0({'a': 0, 'b': 1, 'c': 2, 'd': 3, 'e': 4})
    assert not filter_0({'b': 1, 'c': 2, 'd': 3})
    assert not filter_0({'a': 1, 'b': 1, 'c': 2})

    filter_1 = neurovault.ResultFilter(query_terms={'c': 2})
    filter_1['d'] = neurovault.NotNull()
    assert filter_1({'c': 2, 'd': 1})
    assert not filter_1({'c': 2, 'd': 0})
    filter_1['d'] = neurovault.IsIn(0, 1)
    assert filter_1({'c': 2, 'd': 1})
    assert not filter_1({'c': 2, 'd': 2})
    del filter_1['d']
    assert filter_1({'c': 2, 'd': 2})
    filter_1['d'] = neurovault.NotIn(0, 1)
    assert not filter_1({'c': 2, 'd': 1})
    assert filter_1({'c': 2, 'd': 3})
    filter_1.add_filter(lambda d: len(d) > 2)
    assert not filter_1({'c': 2, 'd': 3})
    assert filter_1({'c': 2, 'd': 3, 'e': 4})


def test_result_filter_combinations():
    filter_0 = neurovault.ResultFilter(a=0, b=1)
    filter_1 = neurovault.ResultFilter(c=2, d=3)

    filter_0_and_1 = filter_0.AND(filter_1)
    assert filter_0_and_1({'a': 0, 'b': 1, 'c': 2, 'd': 3})
    assert not filter_0_and_1({'a': 0, 'b': 1, 'c': 2, 'd': None})
    assert not filter_0_and_1({'a': None, 'b': 1, 'c': 2, 'd': 3})

    filter_0_or_1 = filter_0.OR(filter_1)
    assert filter_0_or_1({'a': 0, 'b': 1, 'c': 2, 'd': 3})
    assert filter_0_or_1({'a': 0, 'b': 1, 'c': 2, 'd': None})
    assert filter_0_or_1({'a': None, 'b': 1, 'c': 2, 'd': 3})
    assert not filter_0_or_1({'a': None, 'b': 1, 'c': 2, 'd': None})

    filter_0_xor_1 = filter_0.XOR(filter_1)
    assert not filter_0_xor_1({'a': 0, 'b': 1, 'c': 2, 'd': 3})
    assert filter_0_xor_1({'a': 0, 'b': 1, 'c': 2, 'd': None})
    assert filter_0_xor_1({'a': None, 'b': 1, 'c': 2, 'd': 3})
    assert not filter_0_xor_1({'a': None, 'b': 1, 'c': 2, 'd': None})

    not_filter_0 = filter_0.NOT()
    assert not_filter_0({})
    assert not not_filter_0({'a': 0, 'b': 1})

    filter_2 = neurovault.ResultFilter(
        {'a': neurovault.NotNull()}).AND(lambda d: len(d) < 2)
    assert filter_2({'a': 'a'})
    assert not filter_2({'a': ''})
    assert not filter_2({'a': 'a', 'b': 0})

    filt = neurovault.ResultFilter(
        a=0).AND(neurovault.ResultFilter(b=1).OR(neurovault.ResultFilter(b=2)))
    assert filt({'a': 0, 'b': 1})
    assert not filt({'a': 0, 'b': 0})


def test_simple_download(request_mocker):
    with tempfile.TemporaryDirectory() as temp_dir:
        downloaded_file = neurovault._simple_download(
            'http://neurovault.org/media/images/35/Fig3B_zstat1.nii.gz',
            os.path.join(temp_dir, 'image_35.nii.gz'), temp_dir)
        assert os.path.isfile(downloaded_file)
        request_mocker.url_mapping["*"] = requests.RequestException()
        pytest.raises(requests.RequestException,
                      neurovault._simple_download, 'http://',
                      os.path.join(temp_dir, 'bad.nii.gz'), temp_dir)


def test_neurosynth_words_vectorized():
    n_im = 5
    with tempfile.TemporaryDirectory() as temp_dir:
        words_files = [
            os.path.join(temp_dir, 'words_for_image_{0}.json'.format(i)) for
            i in range(n_im)]
        words = [str(i) for i in range(n_im)]
        for i, file_name in enumerate(words_files):
            word_weights = np.zeros(n_im)
            word_weights[i] = 1
            words_dict = {'data':
                          {'values':
                           dict([(k, v) for
                                 k, v in zip(words, word_weights)])}}
            with open(file_name, 'wb') as words_file:
                words_file.write(json.dumps(words_dict).encode('utf-8'))
        freq, voc = neurovault.neurosynth_words_vectorized(words_files)
        assert freq.shape == (n_im, n_im)
        assert((freq.sum(axis=0) == np.ones(n_im)).all())
        with pytest.warns(UserWarning):
            neurovault.neurosynth_words_vectorized(
                (os.path.join(temp_dir, 'no_words_here.json'),)
            )


def test_write_read_metadata():
    metadata = {'relative_path': 'collection_1',
                'absolute_path': os.path.join('tmp', 'collection_1')}
    with tempfile.TemporaryDirectory() as temp_dir:
        neurovault._write_metadata(
            metadata, os.path.join(temp_dir, 'metadata.json'))
        with open(os.path.join(temp_dir, 'metadata.json'), 'rb') as meta_file:
            written_metadata = json.loads(meta_file.read().decode('utf-8'))
        assert 'relative_path' in written_metadata
        assert not 'absolute_path' in written_metadata
        read_metadata = neurovault._add_absolute_paths('tmp', written_metadata)
        assert (read_metadata['absolute_path'] ==
                     os.path.join('tmp', 'collection_1'))


def test_add_absolute_paths():
    meta = {'col_relative_path': 'collection_1',
            'col_absolute_path': os.path.join(
                'dir_0', 'neurovault', 'collection_1')}
    meta = neurovault._add_absolute_paths(os.path.join('dir_1', 'neurovault'),
                                          meta, force=False)
    assert (meta['col_absolute_path'] ==
                 os.path.join('dir_0', 'neurovault', 'collection_1'))
    meta = neurovault._add_absolute_paths(os.path.join('dir_1', 'neurovault'),
                                          meta, force=True)
    assert (meta['col_absolute_path'] ==
                 os.path.join('dir_1', 'neurovault', 'collection_1'))
    meta = {'id': 0}
    meta_transformed = neurovault._add_absolute_paths(
        os.path.join('dir_1', 'neurovault'), meta, force=True)
    assert meta == meta_transformed


def test_json_add_collection_dir():
    with tempfile.TemporaryDirectory() as data_temp_dir:
        coll_dir = os.path.join(data_temp_dir, 'collection_1')
        os.makedirs(coll_dir)
        coll_file_name = os.path.join(coll_dir, 'collection_1.json')
        with open(coll_file_name, 'wb') as coll_file:
            coll_file.write(json.dumps({'id': 1}).encode('utf-8'))
        loaded = neurovault._json_add_collection_dir(coll_file_name)
        assert loaded['absolute_path'] == coll_dir
        assert loaded['relative_path'] == 'collection_1'


def test_json_add_im_files_paths():
    with tempfile.TemporaryDirectory() as data_temp_dir:
        coll_dir = os.path.join(data_temp_dir, 'collection_1')
        os.makedirs(coll_dir)
        im_file_name = os.path.join(coll_dir, 'image_1.json')
        with open(im_file_name, 'wb') as im_file:
            im_file.write(json.dumps({'id': 1}).encode('utf-8'))
        loaded = neurovault._json_add_im_files_paths(im_file_name)
        assert (loaded['relative_path'] ==
                     os.path.join('collection_1', 'image_1.nii.gz'))
        assert loaded.get('neurosynth_words_relative_path') is None


def test_split_terms():
    terms, server_terms = neurovault._split_terms(
        {'DOI': neurovault.NotNull(),
         'name': 'my_name', 'unknown_term': 'something'},
        neurovault._COL_FILTERS_AVAILABLE_ON_SERVER)
    assert (terms ==
                 {'DOI': neurovault.NotNull(), 'unknown_term': 'something'})
    assert server_terms == {'name': 'my_name'}


def test_move_unknown_terms_to_local_filter():
    terms, new_filter = neurovault._move_unknown_terms_to_local_filter(
        {'a': 0, 'b': 1}, neurovault.ResultFilter(), ('a',))
    assert terms == {'a': 0}
    assert not new_filter({'b': 0})
    assert new_filter({'b': 1})


def test_move_col_id():
    im_terms, col_terms = neurovault._move_col_id(
        {'collection_id': 1, 'not_mni': False}, {})
    assert im_terms == {'not_mni': False}
    assert col_terms == {'id': 1}

    with pytest.warns(UserWarning):
        neurovault._move_col_id(
            {'collection_id': 1, 'not_mni': False}, {'id': 2}
        )


def test_download_image_terms(request_mocker):
    with tempfile.TemporaryDirectory() as temp_dir:
        image_info = {'id': 'a'}
        collection = {'relative_path': 'collection',
                      'absolute_path': os.path.join(temp_dir, 'collection')}
        os.makedirs(collection['absolute_path'])
        download_params = {'temp_dir': temp_dir, 'verbose': 3,
                           'fetch_neurosynth_words': True}
        request_mocker.url_mapping["*"] = requests.RequestException()
        neurovault._download_image_terms(
            image_info, collection, download_params)
        download_params['allow_neurosynth_failure'] = False
        pytest.raises(RuntimeError,
                      neurovault._download_image_terms,
                      image_info, collection, download_params)
        with open(os.path.join(
                collection['absolute_path'],
                'neurosynth_words_for_image_a.json'), 'w'):
            pass
        neurovault._download_image_terms(
            image_info, collection, download_params)


def test_download_image():
    image = neurovault._download_image(None, {})
    assert image is None


def test_fetch_neurovault(tmp_path):
    # check that nothing is downloaded in offline mode
    data = neurovault.fetch_neurovault(
        mode='offline', data_dir=str(tmp_path))
    assert len(data.images) == 0
    # try to download an image
    data = neurovault.fetch_neurovault(
        max_images=11, fetch_neurosynth_words=True,
        mode='overwrite', data_dir=str(tmp_path))
    # specifying a filter while leaving the default term
    # filters in place should raise a warning.
    with pytest.warns(UserWarning):
        neurovault.fetch_neurovault(
            image_filter=lambda x: True, max_images=1,
            mode='offline'
        )

    assert data.images
    assert len(data.images) == 11
    for meta in data.images_meta:
        assert not meta['not_mni']
        assert not meta['is_thresholded']
        assert not meta['map_type'] in [
            'ROI/mask', 'anatomical', 'parcellation']
        assert not meta['image_type'] == 'atlas'

    # using a data directory we can't write into should raise a
    # warning unless mode is 'offline'
    os.chmod(str(tmp_path), stat.S_IREAD | stat.S_IEXEC)
    os.chmod(os.path.join(str(tmp_path), 'neurovault'),
             stat.S_IREAD | stat.S_IEXEC)
    if os.access(os.path.join(str(tmp_path), 'neurovault'), os.W_OK):
        return
    with pytest.warns(UserWarning):
        neurovault.fetch_neurovault(data_dir=str(tmp_path))


def test_fetch_neurovault_errors(request_mocker):
    request_mocker.url_mapping["*"] = 500
    data = neurovault.fetch_neurovault()
    assert len(data.images) == 0


def test_fetch_neurovault_ids(tmp_path):
    data_dir = str(tmp_path)
    collections, images = _get_neurovault_data()
    collections = collections.sort_values(
        by="true_number_of_images", ascending=False)
    other_col_id, *col_ids = collections["id"].values[:3]
    img_ids = images[images["collection_id"] == other_col_id]["id"].values[:3]
    img_from_cols_ids = images[
        images["collection_id"].isin(col_ids)]["id"].values
    pytest.raises(ValueError, neurovault.fetch_neurovault_ids, mode='bad')
    data = neurovault.fetch_neurovault_ids(
        image_ids=img_ids, collection_ids=col_ids, data_dir=data_dir)
    expected_images = list(img_ids) + list(img_from_cols_ids)
    assert len(data.images) == len(expected_images)
    assert {img['id'] for img in data['images_meta']} == set(expected_images)
    assert os.path.dirname(
        data['images'][0]) == data['collections_meta'][0]['absolute_path']
    # check image can be loaded again from disk
    data = neurovault.fetch_neurovault_ids(
        image_ids=[img_ids[0]], data_dir=data_dir, mode='offline')
    assert len(data.images) == 1
    # check that download_new mode forces overwrite
    modified_meta = data['images_meta'][0]
    assert modified_meta['some_key'] == 'some_value'
    modified_meta['some_key'] = 'some_other_value'
    # mess it up on disk
    meta_path = os.path.join(
        os.path.dirname(modified_meta['absolute_path']),
        'image_{}_metadata.json'.format(img_ids[0]))
    with open(meta_path, 'wb') as meta_f:
        meta_f.write(json.dumps(modified_meta).encode('UTF-8'))
    # fresh download
    data = neurovault.fetch_neurovault_ids(
        image_ids=[img_ids[0]], data_dir=data_dir, mode='download_new')
    data = neurovault.fetch_neurovault_ids(
        image_ids=[img_ids[0]], data_dir=data_dir, mode='offline')
    # should not have changed
    assert data['images_meta'][0]['some_key'] == 'some_other_value'
    data = neurovault.fetch_neurovault_ids(
        image_ids=[img_ids[0]], data_dir=data_dir, mode='overwrite')
    data = neurovault.fetch_neurovault_ids(
        image_ids=[img_ids[0]], data_dir=data_dir, mode='offline')
    # should be back to the original version
    assert data['images_meta'][0]['some_key'] == 'some_value'


def test_should_download_resampled_images_only_if_no_previous_download(tmp_path):
    collections, images = _get_neurovault_data()

    sample_collection = collections.iloc[0]
    sample_collection_id = sample_collection["id"]
    expected_number_of_images = sample_collection["true_number_of_images"]

    data = neurovault.fetch_neurovault_ids(
        collection_ids=[sample_collection_id],
        data_dir=str(tmp_path),
        resample=True,
    )

    # Check the expected size of the dataset
    assert (len(data['images_meta'])) == expected_number_of_images

    # Check that the resampled version is here
    assert np.all([os.path.isfile(im_meta['resampled_absolute_path']) for im_meta in data['images_meta']])

    # Load images that are fetched and check the affines
    affines = [load_img(cur_im).affine for cur_im in data['images']]
    assert np.all([np.all(affine == neurovault.STD_AFFINE) for affine in affines])

    # Check that the original version is NOT here
    assert not np.any([os.path.isfile(im_meta['absolute_path']) for im_meta in data['images_meta']])


def test_should_download_original_images_along_resampled_images_if_previously_downloaded(tmp_path):
    collections, images = _get_neurovault_data()

    sample_collection = collections.iloc[0]
    sample_collection_id = sample_collection["id"]

    # Fetch non-resampled images
    data = neurovault.fetch_neurovault_ids(collection_ids=[sample_collection_id], data_dir=str(tmp_path),
                                                resample=True)

    # Check that only the resampled version is here
    assert np.all([os.path.isfile(im_meta['resampled_absolute_path']) for im_meta in data['images_meta']])
    assert not np.any([os.path.isfile(im_meta['absolute_path']) for im_meta in data['images_meta']])

    # Get the time of the last access to the resampled data
    access_time_resampled = (os.path.getatime(data['images_meta'][0]['resampled_absolute_path']))

    # Download original data
    data_orig = neurovault.fetch_neurovault_ids(collection_ids=[sample_collection_id], data_dir=str(tmp_path), resample=False)

    # Get the time of the last access to one of the original files (which should be download time)
    access_time = (os.path.getatime(data_orig['images_meta'][0]['absolute_path']))

    # Check that the last access to the original data is after the access to the resampled data
    assert (access_time - access_time_resampled > 0)

    # Check that the original version is now here (previous test should have failed anyway if not)
    assert np.all([os.path.isfile(im_meta['absolute_path']) for im_meta in data_orig['images_meta']])

    # Check that the affines of the original version do not correspond to the resampled one
    affines_orig = [load_img(cur_im).affine for cur_im in data_orig['images']]
    assert not np.any([np.all(affine == neurovault.STD_AFFINE) for affine in affines_orig])



def test_should_download_resampled_images_along_original_images_if_previously_downloaded(tmp_path):
    collections, images = _get_neurovault_data()

    sample_collection = collections.iloc[0]
    sample_collection_id = sample_collection["id"]

    # Fetch non-resampled images
    data_orig = neurovault.fetch_neurovault_ids(collection_ids=[sample_collection_id], data_dir=str(tmp_path), resample=False)

    # Check that the original version is here
    assert np.all([os.path.isfile(im_meta['absolute_path']) for im_meta in data_orig['images_meta']])

    # Check that the resampled version is NOT here
    assert not np.any([os.path.isfile(im_meta['resampled_absolute_path']) for im_meta in data_orig['images_meta']])

    # Asks for the resampled version. This should only resample, not download.

    # Get the time of the last modification to the original data
    modif_time_original = (os.path.getmtime(data_orig['images_meta'][0]['absolute_path']))

    # Ask for resampled data, which should only trigger resample
    data = neurovault.fetch_neurovault_ids(collection_ids=[sample_collection_id], data_dir=str(tmp_path), resample=True)

    # Get the time of the last modification to the original data, after fetch
    modif_time_original_after = (os.path.getmtime(data['images_meta'][0]['absolute_path']))

    # The time difference should be 0
    assert (np.isclose(modif_time_original, modif_time_original_after))

    # Check that the resampled version is here
    assert np.all([os.path.isfile(im_meta['resampled_absolute_path']) for im_meta in data['images_meta']])

    # And the original version should still be here as well
    assert np.all([os.path.isfile(im_meta['absolute_path']) for im_meta in data['images_meta']])

    # Load resampled images and check the affines
    affines = [load_img(cur_im).affine for cur_im in data['images']]
    assert np.all([np.all(affine == neurovault.STD_AFFINE) for affine in affines])

    # Check that the affines of the original version do not correspond to the resampled one
    affines_orig = [load_img(cur_im).affine for cur_im in data_orig['images']]
    assert not np.any([np.all(affine == neurovault.STD_AFFINE) for affine in affines_orig])



"""
Simple code to simulate data
"""

import numpy as np
from scipy import linalg, ndimage
from sklearn.utils import check_random_state


def create_graph_net_simulation_data(
        snr=1., n_samples=200, size=8, n_points=10, random_state=42,
        task="regression", smooth_X=1):
    """
    Function to generate data

    """
    generator = check_random_state(random_state)
    # Coefs
    w = np.zeros((size, size, size))
    for _ in range(n_points):
        point = (generator.randint(0, size), generator.randint(0, size),
                 generator.randint(0, size))
        w[point] = 1.0
    mask = np.ones((size, size, size), dtype=bool)
    w = ndimage.gaussian_filter(w, sigma=1)
    w = w[mask]

    # Generate smooth background noise
    XX = generator.randn(n_samples, size, size, size)
    noise = []
    for i in range(n_samples):
        Xi = ndimage.filters.gaussian_filter(XX[i, :, :, :], smooth_X)
        Xi = Xi[mask]
        noise.append(Xi)
    noise = np.array(noise)

    # Generate the signal y
    if task == "regression":
        y = generator.randn(n_samples)
    elif task == "classification":
        y = np.ones(n_samples)
        y[0::2] = -1
    X = np.dot(y[:, np.newaxis], w[np.newaxis])
    norm_noise = linalg.norm(X, 2) / np.exp(snr / 20.)
    noise_coef = norm_noise / linalg.norm(noise, 2)
    noise *= noise_coef
    snr = 20 * np.log(linalg.norm(X, 2) / linalg.norm(noise, 2))

    # Mixing of signal + noise and splitting into train/test
    X += noise
    X -= X.mean(axis=-1)[:, np.newaxis]
    X /= X.std(axis=-1)[:, np.newaxis]

    return X, y, w, mask



from sklearn.base import BaseEstimator, TransformerMixin
from nilearn._utils import CacheMixin


class BaseGLM(BaseEstimator, TransformerMixin, CacheMixin):
    def generate_report(
            self, contrasts, title=None, bg_img="MNI152TEMPLATE",
            threshold=3.09, alpha=0.001, cluster_threshold=0,
            height_control='fpr', min_distance=8., plot_type='slice',
            display_mode=None, report_dims=(1600, 800)):
        """ Returns HTMLDocument object
        for a report which shows all important aspects of a fitted GLM.
        The object can be opened in a browser, displayed in a notebook,
        or saved to disk as a standalone HTML file.

        The GLM must be fitted and have the computed design matrix(ces).

        Parameters
        ----------
            A fitted first or second level model object.

        contrasts : Dict[string, ndarray] or String or List[String] or ndarray or
            List[ndarray]

            Contrasts information for a first or second level model.

            Example:

                Dict of contrast names and coefficients,
                or list of contrast names
                or list of contrast coefficients
                or contrast name
                or contrast coefficient

                Each contrast name must be a string.
                Each contrast coefficient must be a list or numpy array of ints.

            Contrasts are passed to ``contrast_def`` for FirstLevelModel
            (:func:`nilearn.glm.first_level.FirstLevelModel.compute_contrast`)
            & second_level_contrast for SecondLevelModel
            (:func:`nilearn.glm.second_level.SecondLevelModel.compute_contrast`)

        title : String, optional
            If string, represents the web page's title and primary heading,
            model type is sub-heading.
            If None, page titles and headings are autogenerated
            using contrast names.

        bg_img : Niimg-like object, optional
            Default is the MNI152 template (Default='MNI152TEMPLATE')
            See http://nilearn.github.io/manipulating_images/input_output.html
            The background image for mask and stat maps to be plotted on upon.
            To turn off background image, just pass "bg_img=None".

        threshold : float, optional
            Cluster forming threshold in same scale as `stat_img` (either a
            t-scale or z-scale value). Used only if height_control is None.
            Default=3.09

        alpha : float, optional
            Number controlling the thresholding (either a p-value or q-value).
            Its actual meaning depends on the height_control parameter.
            This function translates alpha to a z-scale threshold.
            Default=0.001

        cluster_threshold : int, optional
            Cluster size threshold, in voxels.
            Default=0

        height_control : string or None, optional
            false positive control meaning of cluster forming
            threshold: 'fpr' (default) or 'fdr' or 'bonferroni' or None
            Default='fpr'.

        min_distance : float, optional
            For display purposes only.
            Minimum distance between subpeaks in mm. Default=8mm.

        plot_type : String. ['slice', 'glass'], optional
            Specifies the type of plot to be drawn for the statistical maps.
            Default='slice'.

        display_mode : string, optional
            Default is 'z' if plot_type is 'slice'; '
            ortho' if plot_type is 'glass'.

            Choose the direction of the cuts:
            'x' - sagittal, 'y' - coronal, 'z' - axial,
            'l' - sagittal left hemisphere only,
            'r' - sagittal right hemisphere only,
            'ortho' - three cuts are performed in orthogonal directions.

            Possible values are:
            'ortho', 'x', 'y', 'z', 'xz', 'yx', 'yz',
            'l', 'r', 'lr', 'lzr', 'lyr', 'lzry', 'lyrz'.

        report_dims : Sequence[int, int], optional
            Default is (1600, 800) pixels.
            Specifies width, height (in pixels) of report window within a notebook.
            Only applicable when inserting the report into a Jupyter notebook.
            Can be set after report creation using report.width, report.height.

        Returns
        -------
        report_text : HTMLDocument Object
            Contains the HTML code for the GLM Report.

        """
        from nilearn.reporting import make_glm_report
        return make_glm_report(
            self, contrasts, title=title, bg_img=bg_img, threshold=threshold,
            alpha=alpha, cluster_threshold=cluster_threshold,
            height_control=height_control, min_distance=min_distance,
            plot_type=plot_type, display_mode=display_mode,
            report_dims=report_dims)



"""
Test the nifti_masker module

Functions in this file only test features added by the NiftiMasker class,
not the underlying functions used (e.g. clean()). See test_masking.py and
test_signal.py for this.
"""
# Author: Gael Varoquaux, Philippe Gervais
# License: simplified BSD
import os
import shutil
from tempfile import mkdtemp

import nibabel
import numpy as np
import pytest

from nibabel import Nifti1Image
from numpy.testing import assert_array_equal

from nilearn._utils import testing
from nilearn._utils import data_gen
from nilearn._utils.class_inspect import get_params
from nilearn._utils.exceptions import DimensionError
from nilearn.image import index_img
from nilearn.input_data.nifti_masker import NiftiMasker, filter_and_mask
from nilearn.image import get_data


def test_auto_mask():
    # This mostly a smoke test
    data = np.zeros((9, 9, 9))
    data[3:-3, 3:-3, 3:-3] = 10
    img = Nifti1Image(data, np.eye(4))
    masker = NiftiMasker()
    # Smoke test the fit
    masker.fit(img)
    # Smoke test the transform
    # With a 4D img
    masker.transform([img, ])
    # With a 3D img
    masker.transform(img)

    # check exception when transform() called without prior fit()
    masker2 = NiftiMasker(mask_img=img)
    with pytest.raises(ValueError, match='has not been fitted. '):
        masker2.transform(img)


def test_detrend():
    # Check that detrending doesn't do something stupid with 3D images
    data = np.zeros((9, 9, 9))
    data[3:-3, 3:-3, 3:-3] = 10
    img = Nifti1Image(data, np.eye(4))
    mask = data.astype(np.int)
    mask_img = Nifti1Image(mask, np.eye(4))
    masker = NiftiMasker(mask_img=mask_img, detrend=True)
    # Smoke test the fit
    X = masker.fit_transform(img)
    assert np.any(X != 0)


def test_resample():
    # Check that target_affine triggers the right resampling
    data = np.zeros((9, 9, 9))
    data[3:-3, 3:-3, 3:-3] = 10
    img = Nifti1Image(data, np.eye(4))
    mask = data.astype(np.int)
    mask_img = Nifti1Image(mask, np.eye(4))
    masker = NiftiMasker(mask_img=mask_img, target_affine=2 * np.eye(3))
    # Smoke test the fit
    X = masker.fit_transform(img)
    assert np.any(X != 0)


def test_with_files():
    # Standard masking
    data = np.zeros((40, 40, 40, 2))
    data[20, 20, 20] = 1
    data_img = Nifti1Image(data, np.eye(4))

    with testing.write_tmp_imgs(data_img) as filename:
        masker = NiftiMasker()
        masker.fit(filename)
        masker.transform(filename)


def test_nan():
    data = np.ones((9, 9, 9))
    data[0] = np.nan
    data[:, 0] = np.nan
    data[:, :, 0] = np.nan
    data[-1] = np.nan
    data[:, -1] = np.nan
    data[:, :, -1] = np.nan
    data[3:-3, 3:-3, 3:-3] = 10
    img = Nifti1Image(data, np.eye(4))
    masker = NiftiMasker(mask_args=dict(opening=0))
    masker.fit(img)
    mask = get_data(masker.mask_img_)
    assert mask[1:-1, 1:-1, 1:-1].all()
    assert not mask[0].any()
    assert not mask[:, 0].any()
    assert not mask[:, :, 0].any()
    assert not mask[-1].any()
    assert not mask[:, -1].any()
    assert not mask[:, :, -1].any()


def test_matrix_orientation():
    """Test if processing is performed along the correct axis."""

    # the "step" kind generate heavyside-like signals for each voxel.
    # all signals being identical, standardizing along the wrong axis
    # would leave a null signal. Along the correct axis, the step remains.
    fmri, mask = data_gen.generate_fake_fmri(shape=(40, 41, 42), kind="step")
    masker = NiftiMasker(mask_img=mask, standardize=True, detrend=True)
    timeseries = masker.fit_transform(fmri)
    assert(timeseries.shape[0] == fmri.shape[3])
    assert(timeseries.shape[1] == get_data(mask).sum())
    std = timeseries.std(axis=0)
    assert(std.shape[0] == timeseries.shape[1])  # paranoid
    assert(not np.any(std < 0.1))

    # Test inverse transform
    masker = NiftiMasker(mask_img=mask, standardize=False, detrend=False)
    masker.fit()
    timeseries = masker.transform(fmri)
    recovered = masker.inverse_transform(timeseries)
    np.testing.assert_array_almost_equal(get_data(recovered), get_data(fmri))


def test_mask_3d():
    # Dummy mask
    data = np.zeros((40, 40, 40, 2))
    data[20, 20, 20] = 1
    data_img = Nifti1Image(data, np.eye(4))

    with testing.write_tmp_imgs(data_img, create_files=True)\
            as filename:
        masker = NiftiMasker(mask_img=filename)
        pytest.raises(TypeError, masker.fit)


def test_mask_4d():
    # Dummy mask
    mask = np.zeros((10, 10, 10), dtype=int)
    mask[3:7, 3:7, 3:7] = 1
    mask_bool = mask.astype(bool)
    mask_img = Nifti1Image(mask, np.eye(4))

    # Dummy data
    data = np.zeros((10, 10, 10, 5), dtype=int)
    data[..., 0] = 1
    data[..., 1] = 2
    data[..., 2] = 3
    data_img_4d = Nifti1Image(data, np.eye(4))
    data_imgs = [index_img(data_img_4d, 0), index_img(data_img_4d, 1),
                 index_img(data_img_4d, 2)]

    # check whether transform is indeed selecting niimgs subset
    sample_mask = np.array([0, 2])
    masker = NiftiMasker(mask_img=mask_img)
    masker.fit()
    data_trans = masker.transform(data_imgs, sample_mask=sample_mask)
    data_trans_img = index_img(data_img_4d, sample_mask)
    data_trans_direct = get_data(data_trans_img)[mask_bool, :]
    data_trans_direct = np.swapaxes(data_trans_direct, 0, 1)
    assert_array_equal(data_trans, data_trans_direct)

    masker = NiftiMasker(mask_img=mask_img)
    masker.fit()
    data_trans2 = masker.transform(data_img_4d, sample_mask=sample_mask)
    assert_array_equal(data_trans2, data_trans_direct)

    # check deprecation warning, and the old API should still work
    with pytest.warns(DeprecationWarning) as record:
        masker = NiftiMasker(mask_img=mask_img, sample_mask=sample_mask)
        masker.fit()
        data_trans_depr = masker.transform(data_img_4d)
    assert "sample_mask will be removed" in record[0].message.args[0]
    assert_array_equal(data_trans_depr, data_trans_direct)

    # show warning when supplying both, use the sample_mask from transform
    diff_sample_mask = np.array([2, 4])
    data_trans_img_diff = index_img(data_img_4d, diff_sample_mask)
    data_trans_direct_diff = get_data(data_trans_img_diff)[mask_bool, :]
    data_trans_direct_diff = np.swapaxes(data_trans_direct_diff, 0, 1)
    masker = NiftiMasker(mask_img=mask_img, sample_mask=sample_mask)
    masker.fit()
    with pytest.warns(UserWarning, match=r'^Overwriting') as record:
        data_trans3 = masker.transform(data_img_4d,
                                       sample_mask=diff_sample_mask)
    assert_array_equal(data_trans3, data_trans_direct_diff)
    assert "Overwriting deprecated attribute " in record[0].message.args[0]

def test_4d_single_scan():
    mask = np.zeros((10, 10, 10))
    mask[3:7, 3:7, 3:7] = 1
    mask_img = Nifti1Image(mask, np.eye(4))

    # Test that, in list of 4d images with last dimension=1, they are
    # considered as 3d

    rng = np.random.RandomState(42)
    data_5d = [rng.random_sample((10, 10, 10, 1)) for i in range(5)]
    data_4d = [d[..., 0] for d in data_5d]
    data_5d = [nibabel.Nifti1Image(d, np.eye(4)) for d in data_5d]
    data_4d = [nibabel.Nifti1Image(d, np.eye(4)) for d in data_4d]

    masker = NiftiMasker(mask_img=mask_img)
    masker.fit()
    data_trans_5d = masker.transform(data_5d)
    data_trans_4d = masker.transform(data_4d)

    assert_array_equal(data_trans_4d, data_trans_5d)


def test_5d():
    mask = np.zeros((10, 10, 10))
    mask[3:7, 3:7, 3:7] = 1
    mask_img = Nifti1Image(mask, np.eye(4))

    # Test that, in list of 4d images with last dimension=1, they are
    # considered as 3d

    rng = np.random.RandomState(42)
    data_5d = [rng.random_sample((10, 10, 10, 3)) for i in range(5)]
    data_5d = [nibabel.Nifti1Image(d, np.eye(4)) for d in data_5d]

    masker = NiftiMasker(mask_img=mask_img)
    masker.fit()
    with pytest.raises(
            DimensionError,
            match="Input data has incompatible dimensionality: "
                  "Expected dimension is 4D and you provided "
                  "a list of 4D images \\(5D\\)."):
        masker.transform(data_5d)


def test_sessions():
    # Test the sessions vector
    data = np.ones((40, 40, 40, 4))
    # Create a border, so that the masking work well
    data[0] = 0
    data[-1] = 0
    data[:, -1] = 0
    data[:, 0] = 0
    data[..., -1] = 0
    data[..., 0] = 0
    data[20, 20, 20] = 1
    data_img = Nifti1Image(data, np.eye(4))
    masker = NiftiMasker(sessions=np.ones(3, dtype=np.int))
    pytest.raises(ValueError, masker.fit_transform, data_img)

    # check deprecation warning of attribute sessions
    with pytest.warns(DeprecationWarning) as record:
        masker.sessions
    assert "`sessions` attribute is deprecated" in record[0].message.args[0]


def test_joblib_cache():
    from joblib import hash, Memory
    mask = np.zeros((40, 40, 40))
    mask[20, 20, 20] = 1
    mask_img = Nifti1Image(mask, np.eye(4))

    with testing.write_tmp_imgs(mask_img, create_files=True) as filename:
        masker = NiftiMasker(mask_img=filename)
        masker.fit()
        mask_hash = hash(masker.mask_img_)
        get_data(masker.mask_img_)
        assert mask_hash == hash(masker.mask_img_)

        # Test a tricky issue with memmapped joblib.memory that makes
        # imgs return by inverse_transform impossible to save
        cachedir = mkdtemp()
        try:
            masker.memory = Memory(location=cachedir, mmap_mode='r',
                                   verbose=0)
            X = masker.transform(mask_img)
            # inverse_transform a first time, so that the result is cached
            out_img = masker.inverse_transform(X)
            out_img = masker.inverse_transform(X)
            out_img.to_filename(os.path.join(cachedir, 'test.nii'))
        finally:
            # enables to delete "filename" on windows
            del masker
            shutil.rmtree(cachedir, ignore_errors=True)


def test_mask_init_errors():
    # Errors that are caught in init
    mask = NiftiMasker(mask_strategy='oops')
    with pytest.raises(
            ValueError,
            match="Unknown value of mask_strategy 'oops'"):
        mask.fit()


def test_compute_epi_mask():
    # Taken from test_masking.py, but used to test that the masker class
    #   is passing parameters appropriately.
    mean_image = np.ones((9, 9, 3))
    mean_image[3:-2, 3:-2, :] = 10
    mean_image[5, 5, :] = 11
    mean_image = Nifti1Image(mean_image.astype(float), np.eye(4))

    masker = NiftiMasker(mask_strategy='epi',
                         mask_args=dict(opening=False))
    masker.fit(mean_image)
    mask1 = masker.mask_img_

    masker2 = NiftiMasker(mask_strategy='epi',
                          mask_args=dict(opening=False, exclude_zeros=True))
    masker2.fit(mean_image)
    mask2 = masker2.mask_img_

    # With an array with no zeros, exclude_zeros should not make
    # any difference
    np.testing.assert_array_equal(get_data(mask1), get_data(mask2))

    # Check that padding with zeros does not change the extracted mask
    mean_image2 = np.zeros((30, 30, 3))
    mean_image2[3:12, 3:12, :] = get_data(mean_image)
    mean_image2 = Nifti1Image(mean_image2, np.eye(4))

    masker3 = NiftiMasker(mask_strategy='epi',
                          mask_args=dict(opening=False, exclude_zeros=True))
    masker3.fit(mean_image2)
    mask3 = masker3.mask_img_
    np.testing.assert_array_equal(get_data(mask1),
                                  get_data(mask3)[3:12, 3:12])

    # However, without exclude_zeros, it does
    masker4 = NiftiMasker(mask_strategy='epi', mask_args=dict(opening=False))
    masker4.fit(mean_image2)
    mask4 = masker4.mask_img_

    assert not np.allclose(get_data(mask1),
                             get_data(mask4)[3:12, 3:12])


def test_compute_brain_mask():
    # Check masker for template masking strategy

    img = np.random.RandomState(42).uniform(size=(9, 9, 5))
    img = Nifti1Image(img, np.eye(4))

    masker = NiftiMasker(mask_strategy='template')

    masker.fit(img)
    mask1 = masker.mask_img_

    masker2 = NiftiMasker(mask_strategy='template',
                          mask_args=dict(threshold=0.))

    masker2.fit(img)
    mask2 = masker2.mask_img_

    mask_ref = np.zeros((9, 9, 5))
    mask_ref[2:7, 2:7, 2] = 1

    np.testing.assert_array_equal(get_data(mask1), mask_ref)
    np.testing.assert_array_equal(get_data(mask2), mask_ref)


def test_filter_and_mask_error():
    data = np.zeros([20, 30, 40, 5])
    mask = np.zeros([20, 30, 40, 2])
    mask[10, 15, 20, :] = 1

    data_img = nibabel.Nifti1Image(data, np.eye(4))
    mask_img = nibabel.Nifti1Image(mask, np.eye(4))

    masker = NiftiMasker()
    params = get_params(NiftiMasker, masker)

    with pytest.raises(
            DimensionError,
            match="Input data has incompatible dimensionality: "
                  "Expected dimension is 3D and you provided "
                  "a 4D image."):
        filter_and_mask(data_img, mask_img, params)


def test_filter_and_mask():
    data = np.zeros([20, 30, 40, 5])
    mask = np.ones([20, 30, 40])

    data_img = nibabel.Nifti1Image(data, np.eye(4))
    mask_img = nibabel.Nifti1Image(mask, np.eye(4))

    masker = NiftiMasker()
    params = get_params(NiftiMasker, masker)

    # Test return_affine = False
    data = filter_and_mask(data_img, mask_img, params)
    assert data.shape == (5, 24000)


def test_dtype():
    data_32 = np.zeros((9, 9, 9), dtype=np.float32)
    data_64 = np.zeros((9, 9, 9), dtype=np.float64)
    data_32[2:-2, 2:-2, 2:-2] = 10
    data_64[2:-2, 2:-2, 2:-2] = 10

    affine_32 = np.eye(4, dtype=np.float32)
    affine_64 = np.eye(4, dtype=np.float64)

    img_32 = Nifti1Image(data_32, affine_32)
    img_64 = Nifti1Image(data_64, affine_64)

    masker_1 = NiftiMasker(dtype='auto')
    assert(masker_1.fit_transform(img_32).dtype == np.float32)
    assert(masker_1.fit_transform(img_64).dtype == np.float32)

    masker_2 = NiftiMasker(dtype='float64')
    assert(masker_2.fit_transform(img_32).dtype == np.float64)
    assert(masker_2.fit_transform(img_64).dtype == np.float64)


def test_standardization():
    rng = np.random.RandomState(42)
    data_shape = (9, 9, 5)
    n_samples = 500

    signals = rng.standard_normal(size=(np.prod(data_shape), n_samples))
    means = rng.standard_normal(size=(np.prod(data_shape), 1)) * 50 + 1000
    signals += means
    img = Nifti1Image(signals.reshape(data_shape + (n_samples,)), np.eye(4))

    mask = Nifti1Image(np.ones(data_shape), np.eye(4))

    # z-score
    masker = NiftiMasker(mask, standardize='zscore')
    trans_signals = masker.fit_transform(img)

    np.testing.assert_almost_equal(trans_signals.mean(0), 0)
    np.testing.assert_almost_equal(trans_signals.std(0), 1)

    # psc
    masker = NiftiMasker(mask, standardize='psc')
    trans_signals = masker.fit_transform(img)

    np.testing.assert_almost_equal(trans_signals.mean(0), 0)
    np.testing.assert_almost_equal(trans_signals,
                                   (signals / signals.mean(1)[:, np.newaxis] *
                                    100 - 100).T)
